# Copyright © 2020 Hoe Hao Cheng
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
# 
# Authors:
#    Hoe Hao Cheng <haochengho12907@gmail.com>
# 

from mako.template import Template
from mako.lookup import TemplateLookup
from os import path
from zink_extensions import Extension,ExtensionRegistry,Version
import sys

# constructor: 
#     Extension(name, alias="", required=False, properties=False, features=False, conditions=None, guard=False)
# The attributes:
#  - required: the generated code debug_prints "ZINK: {name} required!" and
#              returns NULL if the extension is unavailable.
#
#  - properties: enable the detection of extension properties in a physical
#                device in the generated code using vkGetPhysicalDeviceProperties2(),
#                and store the returned properties struct inside
#                `zink_device_info.{alias}_props`.
#                Example: the properties for `VK_EXT_transform_feedback`, is stored in
#                `VkPhysicalDeviceTransformFeedbackPropertiesEXT tf_props`.
#
#  - features: enable the getting extension features in a
#              device. Similar to `properties`, this stores the features
#              struct inside `zink_device_info.{alias}_feats`.
#
#  - conditions: criteria for enabling an extension. This is an array of strings,
#                where each string is a condition, and all conditions have to be true
#                for `zink_device_info.have_{name}` to be true.
#
#                The code generator will replace "$feats" and "$props" with the
#                respective variables, e.g. "$feats.nullDescriptor" becomes 
#                "info->rb2_feats.nullDescriptor" in the final code for VK_EXT_robustness2.
#
#                When empty or None, the extension is enabled when the extensions
#                given by vkEnumerateDeviceExtensionProperties() include the extension.
#
#  - guard: adds a #if defined(`extension_name`)/#endif guard around the code generated for this Extension.
EXTENSIONS = [
    Extension("VK_KHR_maintenance1",
              required=True),
    Extension("VK_KHR_maintenance2"),
    Extension("VK_KHR_maintenance3"),
    Extension("VK_KHR_maintenance4",
              alias="maint4",
              features=True),
    Extension("VK_KHR_maintenance5",
              alias="maint5",
              features=True, properties=True),
    Extension("VK_KHR_maintenance6",
              alias="maint6",
              features=True, properties=True),
    Extension("VK_KHR_maintenance7",
              alias="maint7",
              features=True, properties=True),
    Extension("VK_KHR_maintenance8",
              alias="maint8",
              features=True),
    Extension("VK_KHR_maintenance9",
              alias="maint9",
              features=True, properties=True),
    Extension("VK_KHR_unified_image_layouts", alias="unified_layouts", features=True),
    Extension("VK_KHR_external_memory"),
    Extension("VK_KHR_external_memory_fd"),
    Extension("VK_KHR_vulkan_memory_model"),
    Extension("VK_KHR_workgroup_memory_explicit_layout", alias="explicit_layout", features=True),
    Extension("VK_KHR_pipeline_executable_properties",
              alias="pipestats",
              features=True),
    Extension("VK_KHR_external_semaphore_fd"),
    Extension("VK_KHR_create_renderpass2",
              required=True),
    Extension("VK_KHR_synchronization2",
              alias="sync2",
              features=True),
    Extension("VK_KHR_external_memory_win32"),
    Extension("VK_KHR_external_semaphore_win32"),
    Extension("VK_EXT_external_memory_dma_buf"),
    Extension("VK_KHR_buffer_device_address",
              alias="bda",
              features=True),
    Extension("VK_EXT_external_memory_host", alias="ext_host_mem", properties=True),
    Extension("VK_EXT_queue_family_foreign"),
    Extension("VK_KHR_swapchain_mutable_format"),
    Extension("VK_KHR_incremental_present"),
    Extension("VK_EXT_provoking_vertex",
              alias="pv",
              features=True,
              properties=True,
              conditions=["$feats.provokingVertexLast"]),
    Extension("VK_EXT_shader_viewport_index_layer"),
    Extension("VK_KHR_get_memory_requirements2"),
    Extension("VK_EXT_post_depth_coverage"),
    Extension("VK_EXT_depth_clip_control",
              alias="clip_control",
              features=True),
    Extension("VK_EXT_depth_clamp_zero_one",
              alias="clamp_01",
              features=True),
    Extension("VK_EXT_shader_subgroup_ballot"),
    Extension("VK_EXT_shader_subgroup_vote"),
    Extension("VK_EXT_legacy_vertex_attributes", alias="legacyverts", features=True, properties=True),
    Extension("VK_EXT_shader_atomic_float",
              alias="atomic_float",
              features=True),
    Extension("VK_KHR_shader_atomic_int64",
              alias="atomic_int",
              features=True),
    Extension("VK_KHR_8bit_storage",
              alias="storage_8bit",
              features=True,
              conditions=["$feats.storageBuffer8BitAccess"]),
    Extension("VK_KHR_16bit_storage",
              alias="storage_16bit",
              features=True,
              conditions=["$feats.storageBuffer16BitAccess"]),
    Extension("VK_EXT_image_2d_view_of_3d",
              alias="view2d",
              features=True),
    Extension("VK_KHR_driver_properties",
              alias="driver",
              properties=True),
    Extension("VK_EXT_memory_budget"),
    Extension("VK_EXT_memory_priority", alias="memprio", features=True),
    Extension("VK_EXT_pageable_device_local_memory", alias="mempage", features=True),
    Extension("VK_KHR_draw_indirect_count"),
    Extension("VK_EXT_dynamic_rendering_unused_attachments", alias="unused", features=True),
    Extension("VK_EXT_shader_object", alias="shobj", features=True, properties=True),
    Extension("VK_EXT_attachment_feedback_loop_layout",
              alias="feedback_loop",
              features=True),
    Extension("VK_EXT_attachment_feedback_loop_dynamic_state", alias="feedback_dyn", features=True),
    Extension("VK_EXT_fragment_shader_interlock",
              alias="interlock",
              features=True,
              conditions=["$feats.fragmentShaderSampleInterlock", "$feats.fragmentShaderPixelInterlock"]),
    Extension("VK_EXT_sample_locations",
              alias="sample_locations",
              properties=True),
    Extension("VK_KHR_shader_draw_parameters"),
    Extension("VK_KHR_sampler_mirror_clamp_to_edge"),
    Extension("VK_EXT_descriptor_buffer", alias="db", features=True, properties=True),
    Extension("VK_EXT_conditional_rendering",
              alias="cond_render",
              features=True,
              conditions=["$feats.conditionalRendering"]),
    Extension("VK_EXT_transform_feedback",
              alias="tf",
              properties=True,
              features=True,
              conditions=["$feats.transformFeedback"]),
    Extension("VK_EXT_index_type_uint8",
              alias="index_uint8",
              features=True,
              conditions=["$feats.indexTypeUint8"]),
    Extension("VK_KHR_image_format_list"),
    Extension("VK_KHR_sampler_ycbcr_conversion"),
    Extension("VK_KHR_imageless_framebuffer",
              alias="imgless",
              features=True,
              required=True),
    Extension("VK_EXT_robustness2",
              alias="rb2",
              properties=True,
              features=True,
              conditions=["$feats.nullDescriptor"]),
    Extension("VK_EXT_image_robustness",
              alias="rb_image",
              features=True),
    Extension("VK_EXT_image_drm_format_modifier"),
    Extension("VK_EXT_vertex_attribute_divisor",
              alias="vdiv",
              properties=True,
              features=True,
              conditions=["$feats.vertexAttributeInstanceRateDivisor"]),
    Extension("VK_EXT_calibrated_timestamps"),
    Extension("VK_NV_representative_fragment_test"),
    Extension("VK_NV_linear_color_attachment",
              alias="linear_color",
              features=True),
    Extension("VK_KHR_dynamic_rendering",
              alias="dynamic_render",
              features=True,
              required=True),
    Extension("VK_KHR_dynamic_rendering_local_read",
              alias="drlr",
              features=True),
    Extension("VK_EXT_multisampled_render_to_single_sampled",
              alias="msrtss",
              features=True),
    Extension("VK_KHR_shader_clock",
              alias="shader_clock",
              features=True,
              conditions=["$feats.shaderSubgroupClock"]),
    Extension("VK_INTEL_shader_integer_functions2",
              alias="shader_int_fns2",
              features=True,
              conditions=["$feats.shaderIntegerFunctions2"]),
    Extension("VK_EXT_sampler_filter_minmax",
              alias="reduction",
              properties=True,
              conditions=["$props.filterMinmaxSingleComponentFormats"]),
    Extension("VK_EXT_custom_border_color",
              alias="border_color",
              properties=True,
              features=True,
              conditions=["$feats.customBorderColors"]),
    Extension("VK_EXT_non_seamless_cube_map",
              alias="nonseamless",
              features=True),
    Extension("VK_EXT_border_color_swizzle",
              alias="border_swizzle",
              features=True),
    Extension("VK_EXT_blend_operation_advanced",
              alias="blend",
              properties=True,
              # TODO: we can probably support non-premul here with some work?
              conditions=["$props.advancedBlendNonPremultipliedSrcColor", "$props.advancedBlendNonPremultipliedDstColor"]),
    Extension("VK_EXT_extended_dynamic_state",
              alias="dynamic_state",
              features=True,
              conditions=["$feats.extendedDynamicState"]),
    Extension("VK_EXT_extended_dynamic_state2",
              alias="dynamic_state2",
              features=True,
              conditions=["$feats.extendedDynamicState2"]),
    Extension("VK_EXT_extended_dynamic_state3",
              alias="dynamic_state3",
              properties=True,
              features=True),
    Extension("VK_EXT_pipeline_creation_cache_control",
              alias="pipeline_cache_control",
              features=True,
              conditions=["$feats.pipelineCreationCacheControl"]),
    Extension("VK_EXT_shader_stencil_export",
              alias="stencil_export"),
    Extension("VK_KHR_portability_subset",
              alias="portability_subset",
              features=True,
              guard=True),
    Extension("VK_NV_compute_shader_derivatives",
              alias="shader_derivs",
              features=True,
              conditions=["$feats.computeDerivativeGroupQuads", "$feats.computeDerivativeGroupLinear"]),
    Extension("VK_KHR_timeline_semaphore",
              alias="timeline",
              features=True,
              properties=True),
    Extension("VK_EXT_color_write_enable",
              alias="cwrite",
              features=True),
    Extension("VK_EXT_4444_formats",
              alias="format_4444",
              features=True),
    Extension("VK_EXT_host_image_copy",
              alias="hic",
              features=True,
              properties=True),
    Extension("VK_EXT_mesh_shader",
              alias="mesh",
              features=True,
              properties=True),
    Extension("VK_EXT_scalar_block_layout",
              alias="scalar_block_layout",
              features=True,
              conditions=["$feats.scalarBlockLayout"]),
    Extension("VK_KHR_swapchain"),
    Extension("VK_EXT_rasterization_order_attachment_access",
              alias="rast_order_access",
              features=True,
              conditions=["$feats.rasterizationOrderColorAttachmentAccess"]),
    Extension("VK_KHR_shader_float16_int8",
              alias="shader_float16_int8",
              features=True),
    Extension("VK_EXT_multi_draw",
              alias="multidraw",
              features=True,
              properties=True,
              conditions=["$feats.multiDraw"]),
    Extension("VK_EXT_primitives_generated_query",
              alias="primgen",
              features=True),
    Extension("VK_KHR_pipeline_library"),
    Extension("VK_EXT_graphics_pipeline_library",
              alias="gpl",
              features=True,
              properties=True),
    Extension("VK_KHR_push_descriptor",
              alias="push",
              properties=True),
    Extension("VK_KHR_descriptor_update_template",
              alias="template", required=True),
    Extension("VK_EXT_line_rasterization",
              alias="line_rast",
              properties=True,
              features=True),
    Extension("VK_EXT_vertex_input_dynamic_state",
              alias="vertex_input",
              features=True,
              conditions=["$feats.vertexInputDynamicState"]),
    Extension("VK_EXT_primitive_topology_list_restart",
              alias="list_restart",
              features=True,
              conditions=["$feats.primitiveTopologyListRestart"]),
    Extension("VK_KHR_dedicated_allocation",
              alias="dedicated"),
    Extension("VK_EXT_descriptor_indexing",
              alias="desc_indexing",
              features=True,
              properties=True,
              conditions=["$feats.descriptorBindingPartiallyBound"]),
    Extension("VK_EXT_depth_clip_enable",
              alias="depth_clip_enable",
              features=True),
    Extension("VK_EXT_shader_demote_to_helper_invocation",
              alias="demote",
              features=True,
              conditions=["$feats.shaderDemoteToHelperInvocation"]),
    Extension("VK_KHR_shader_float_controls",
              alias="float_controls"),
    Extension("VK_KHR_format_feature_flags2"),
]

# constructor: Versions(device_version(major, minor, patch), struct_version(major, minor))
# The attributes:
#  - device_version: Vulkan version, as tuple, to use with 
#                    VK_MAKE_VERSION(version_major, version_minor, version_patch)
#
#  - struct_version: Vulkan version, as tuple, to use with structures and macros
VERSIONS = [
    Version((1,1,0), (1,1)),
    Version((1,2,0), (1,2)),
    Version((1,3,0), (1,3)),
    Version((1,4,0), (1,4)),
]

# There exists some inconsistencies regarding the enum constants, fix them.
# This is basically generated_code.replace(key, value).
REPLACEMENTS = {
    "PROPERTIES_PROPERTIES": "PROPERTIES",
}


# This template provides helper functions for the other templates.
# Right now, the following functions are defined:
# - guard(ext) : surrounds the body with an if-def guard according to
#                `ext.extension_name()` if `ext.guard` is True.
include_template = """
<%def name="guard_(ext, body)">
   %if ext.guard:
      #ifdef ${ext.extension_name()}
   %endif
      ${capture(body)|trim}
   %if ext.guard:
      #endif
   %endif
</%def>

## This ugliness is here to prevent mako from adding tons of excessive whitespace
<%def name="guard(ext)">${capture(guard_, ext, body=caller.body).strip('\\r\\n')}</%def>
"""

header_code = """
<%namespace name="helpers" file="helpers"/>

#ifndef ZINK_DEVICE_INFO_H
#define ZINK_DEVICE_INFO_H

#include "util/u_memory.h"

#include <vulkan/vulkan_core.h>

#ifdef VK_ENABLE_BETA_EXTENSIONS
#include <vulkan/vulkan_beta.h>
#endif

#ifdef _WIN32
#include <windows.h>
#include <vulkan/vulkan_win32.h>
#endif

struct zink_screen;

struct zink_device_info {
   uint32_t device_version;

%for ext in extensions:
   <%helpers:guard ext="${ext}">
      bool have_${ext.name_with_vendor()};
   </%helpers:guard>
%endfor
%for version in versions:
   bool have_vulkan${version.struct()};
%endfor

   VkPhysicalDeviceFeatures2 feats;
   VkPhysicalDeviceSubgroupProperties subgroup;
%for version in versions:
   VkPhysicalDeviceVulkan${version.struct()}Features feats${version.struct()};
%endfor

   VkPhysicalDeviceProperties props;
   VkPhysicalDeviceProperties vk_layered_props;
   VkPhysicalDeviceLayeredApiPropertiesKHR layered_props;
   VkPhysicalDeviceDriverPropertiesKHR vk_layered_driver_props;
%for version in versions:
   VkPhysicalDeviceVulkan${version.struct()}Properties props${version.struct()};
%endfor

   VkPhysicalDeviceMemoryProperties mem_props;
   VkPhysicalDeviceIDProperties deviceid_props;

%for ext in extensions:
   <%helpers:guard ext="${ext}">
      %if ext.has_features:
         ${ext.physical_device_struct("Features")} ${ext.field("feats")};
      %endif
      %if ext.has_properties:
         ${ext.physical_device_struct("Properties")} ${ext.field("props")};
      %endif
   </%helpers:guard>
%endfor

    const char *extensions[${len(extensions)}];
    uint32_t num_extensions;
};

bool
zink_get_physical_device_info(struct zink_screen *screen);

void
zink_verify_device_extensions(struct zink_screen *screen);

/* stub functions that get inserted into the dispatch table if they are not
 * properly loaded.
 */
%for ext in extensions:
   %if registry.in_registry(ext.name):
      %for cmd in registry.get_registry_entry(ext.name).device_commands:
         void VKAPI_PTR zink_stub_${cmd.name()}(void);
      %endfor
   %endif
%endfor

#endif
"""


impl_code = """
<%namespace name="helpers" file="helpers"/>

#include "vk_enum_to_str.h"
#include "zink_device_info.h"
#include "zink_screen.h"

bool
zink_get_physical_device_info(struct zink_screen *screen) 
{
   struct zink_device_info *info = &screen->info;
%for ext in extensions:
   <%helpers:guard ext="${ext}">
         bool support_${ext.name_with_vendor()} = false;
      %if ext.is_promoted_to_khr:
         bool support_${ext.name_with_vendor("KHR")} = false; /* promoted from EXT */
      %endif
   </%helpers:guard>
%endfor
   uint32_t num_extensions = 0;

   // get device memory properties
   screen->vk.GetPhysicalDeviceMemoryProperties(screen->pdev, &info->mem_props);

   // enumerate device supported extensions
   VkResult result = screen->vk.EnumerateDeviceExtensionProperties(screen->pdev, NULL, &num_extensions, NULL);
   if (result != VK_SUCCESS) {
      if (!screen->driver_name_is_inferred)
         mesa_loge("ZINK: vkEnumerateDeviceExtensionProperties failed (%s)", vk_Result_to_str(result));
   } else {
      if (num_extensions > 0) {
         VkExtensionProperties *extensions = MALLOC(sizeof(VkExtensionProperties) * num_extensions);
         if (!extensions) goto fail;
         result = screen->vk.EnumerateDeviceExtensionProperties(screen->pdev, NULL, &num_extensions, extensions);
         if (result != VK_SUCCESS) {
            if (!screen->driver_name_is_inferred)
               mesa_loge("ZINK: vkEnumerateDeviceExtensionProperties failed (%s)", vk_Result_to_str(result));
         }

         for (uint32_t i = 0; i < num_extensions; ++i) {
         %for ext in extensions:
            <%helpers:guard ext="${ext}">
               %if ext.is_promoted_to_khr:
                  bool promoted_${ext.pure_name()} = !strcmp(extensions[i].extensionName, "${ext.with_vendor("KHR")}");
                  if (!strcmp(extensions[i].extensionName, "${ext.name}") || promoted_${ext.pure_name()}) {
               %else:
                  if (!strcmp(extensions[i].extensionName, "${ext.name}")) {
               %endif
               %if not (ext.has_features or ext.has_properties):
                     info->have_${ext.name_with_vendor()} = true;
               %else:
                     support_${ext.name_with_vendor()} = true;
               %endif
               %if ext.is_promoted_to_khr:
                     if (promoted_${ext.pure_name()}) {
                        support_${ext.name_with_vendor("KHR")} = true;
                     }
               %endif
                  }
            </%helpers:guard>
         %endfor
         }

         FREE(extensions);
      }
   }

   // get device features
   if (screen->vk.GetPhysicalDeviceFeatures2) {
      // check for device extension features
      info->feats.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2;

%for version in versions:
   %if version.device_version < (1,2,0):
         if (VK_MAKE_VERSION(1,2,0) <= screen->vk_version) {
            /* VkPhysicalDeviceVulkan11Features was added in 1.2, not 1.1 as one would think */
   %else:
         if (${version.version()} <= screen->vk_version) {
   %endif
            info->feats${version.struct()}.sType = ${version.stype("FEATURES")};
            info->feats${version.struct()}.pNext = info->feats.pNext;
            info->feats.pNext = &info->feats${version.struct()};
            info->have_vulkan${version.struct()} = true;
         }
%endfor

%for ext in extensions:
   %if ext.has_features:
      <%helpers:guard ext="${ext}">
         %if ext.features_promoted:
               if (support_${ext.name_with_vendor()} && !info->have_vulkan${ext.core_since.struct()}) {
         %else:
               if (support_${ext.name_with_vendor()}) {
         %endif
                  info->${ext.field("feats")}.sType = ${ext.stype("FEATURES")};
                  info->${ext.field("feats")}.pNext = info->feats.pNext;
                  info->feats.pNext = &info->${ext.field("feats")};
               }
      </%helpers:guard>
   %endif
%endfor

      screen->vk.GetPhysicalDeviceFeatures2(screen->pdev, &info->feats);
   } else {
      screen->vk.GetPhysicalDeviceFeatures(screen->pdev, &info->feats.features);
   }

   // check for device properties
   bool copy_layered_props = false;
   if (screen->vk.GetPhysicalDeviceProperties2) {
      VkPhysicalDeviceProperties2 props = {0};
      props.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2;

%for version in versions:
   %if version.device_version < (1,2,0):
         if (VK_MAKE_VERSION(1,2,0) <= screen->vk_version) {
            /* VkPhysicalDeviceVulkan11Properties was added in 1.2, not 1.1 as one would think */
   %else:
         if (${version.version()} <= screen->vk_version) {
   %endif
            info->props${version.struct()}.sType = ${version.stype("PROPERTIES")};
            info->props${version.struct()}.pNext = props.pNext;
            props.pNext = &info->props${version.struct()};
         }
%endfor

%for ext in extensions:
   %if ext.has_properties:
      <%helpers:guard ext="${ext}">
         %if ext.properties_promoted:
               if (support_${ext.name_with_vendor()} && !info->have_vulkan${ext.core_since.struct()}) {
         %else:
               if (support_${ext.name_with_vendor()}) {
         %endif
                  info->${ext.field("props")}.sType = ${ext.stype("PROPERTIES")};
                  info->${ext.field("props")}.pNext = props.pNext;
                  props.pNext = &info->${ext.field("props")};
               }
      </%helpers:guard>
   %endif
%endfor

      if (screen->vk_version < VK_MAKE_VERSION(1,2,0) && screen->instance_info->have_KHR_external_memory_capabilities) {
         info->deviceid_props.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ID_PROPERTIES;
         info->deviceid_props.pNext = props.pNext;
         props.pNext = &info->deviceid_props;
      }

      if (screen->vk_version >= VK_MAKE_VERSION(1,1,0)) {
         info->subgroup.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_PROPERTIES;
         info->subgroup.pNext = props.pNext;
         props.pNext = &info->subgroup;
      }

      /* set up structs to capture underlying driver info */
      VkPhysicalDeviceLayeredApiVulkanPropertiesKHR vk_layered_props = {
        VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_VULKAN_PROPERTIES_KHR,
      };
      vk_layered_props.properties.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2;
      info->vk_layered_driver_props.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRIVER_PROPERTIES;
      if (support_KHR_driver_properties || info->have_vulkan12)
        vk_layered_props.properties.pNext = &info->vk_layered_driver_props;
      info->layered_props.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_PROPERTIES_KHR;
      info->layered_props.pNext = &vk_layered_props;
      VkPhysicalDeviceLayeredApiPropertiesListKHR layered_props_list = {
        VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_PROPERTIES_LIST_KHR,
        props.pNext,
        1,
        &info->layered_props
      };
      if (support_KHR_maintenance7)
         props.pNext = &layered_props_list;

      // note: setting up local VkPhysicalDeviceProperties2.
      screen->vk.GetPhysicalDeviceProperties2(screen->pdev, &props);

      if (support_KHR_maintenance7 && layered_props_list.layeredApiCount) {
        info->vk_layered_props = vk_layered_props.properties.properties;
      } else {
        info->vk_layered_props = info->props;
        copy_layered_props = true;
      }
   }

   /* We re-apply the fields from VkPhysicalDeviceVulkanXYFeatures struct
    * onto their respective fields in the VkPhysicalDeviceExtensionNameFeatures
    * struct if the former is provided by the VK implementation.
    *
    * As for why this is done: the spec mentions that once an extension is
    * promoted to core and its feature fields are added in VulkanXYFeatures,
    * including both ExtensionNameFeatures and VulkanXYFeatures at the same
    * time is prohibited when using vkGetPhysicalDeviceFeatures2.
    */
%for ext in extensions:
   %if ext.features_promoted:
      if (info->have_vulkan${ext.core_since.struct()}) {
      %for field in registry.get_registry_entry(ext.name).features_fields:
         info->${ext.field("feats")}.${field} = info->feats${ext.core_since.struct()}.${field};
      %endfor
      }
   %endif
%endfor

   /* See above, but for VulkanXYProperties.
    * Unlike VulkanXYFeatures with all the booleans, VulkanXYProperties can
    * contain different types of data, including arrays. The C language hates us
    * when we assign an array to another array, therefore we use an memcpy here.
    */
%for ext in extensions:
   %if ext.properties_promoted:
      if (info->have_vulkan${ext.core_since.struct()}) {
      %for field in registry.get_registry_entry(ext.name).properties_fields:
         memcpy(&info->${ext.field("props")}.${field.name},
               &info->props${ext.core_since.struct()}.${field.name},
               sizeof(info->${ext.field("props")}.${field.name}));
      %endfor
      }
   %endif
%endfor

   if (copy_layered_props)
     info->vk_layered_driver_props = info->driver_props;

   // enable the extensions if they match the conditions given by ext.enable_conds 
   if (screen->vk.GetPhysicalDeviceProperties2) {
        %for ext in extensions:
<%helpers:guard ext="${ext}">
<%
    conditions = ""
    if ext.enable_conds:
        for cond in ext.enable_conds:
            cond = cond.replace("$feats", "info->" + ext.field("feats"))
            cond = cond.replace("$props", "info->" + ext.field("props"))
            conditions += "&& (" + cond + ")\\n"
    conditions = conditions.strip()
%>\
      info->have_${ext.name_with_vendor()} |= support_${ext.name_with_vendor()}
         ${conditions};
</%helpers:guard>
        %endfor
   }

   /* For extensions that require double-loading (first load for getting array
    * length and malloc-ing, second load for getting array content)
    */
   if (screen->vk.GetPhysicalDeviceProperties2) {
      VkPhysicalDeviceProperties2 second_load_props = {
         .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2
      };

   %for ext in extensions:
      %if ext.needs_double_load:
         if (screen->info.have_${ext.name_with_vendor()}) {
            screen->info.${ext.field("props")}.sType = ${ext.stype("PROPERTIES")};
            screen->info.${ext.field("props")}.pNext = second_load_props.pNext;
            second_load_props.pNext = &screen->info.${ext.field("props")};
         %for field in registry.get_registry_entry(ext.name).properties_fields:
            %if field.len_field():
            screen->info.${ext.field("props")}.${field.name} =
               ralloc_array(screen, ${field.type_name}, screen->info.${ext.field("props")}.${field.len_field()});
            %endif
         %endfor
         }
      %endif
   %endfor

      if (second_load_props.pNext)
         screen->vk.GetPhysicalDeviceProperties2(screen->pdev, &second_load_props);
   }

   // generate extension list
   num_extensions = 0;

%for ext in extensions:
   <%helpers:guard ext="${ext}">
         if (info->have_${ext.name_with_vendor()}) {
      %if ext.is_promoted_to_khr:
            if (support_${ext.name_with_vendor("KHR")})
               info->extensions[num_extensions++] = "${ext.with_vendor("KHR")}";
            else
               info->extensions[num_extensions++] = "${ext.name}";
      %else:
            info->extensions[num_extensions++] = "${ext.name}";
      %endif
      %if ext.is_required:
         } else {
            debug_printf("ZINK: ${ext.name} required!\\n");
            goto fail;
      %endif
         }
   </%helpers:guard>
%endfor

   info->num_extensions = num_extensions;

   info->feats.pNext = NULL;

%for version in versions:
   %if version.device_version < (1,2,0):
         if (VK_MAKE_VERSION(1,2,0) <= screen->vk_version) {
            /* VkPhysicalDeviceVulkan11Features was added in 1.2, not 1.1 as one would think */
   %else:
         if (${version.version()} <= screen->vk_version) {
   %endif
            info->feats${version.struct()}.pNext = info->feats.pNext;
            info->feats.pNext = &info->feats${version.struct()};
         }
%endfor

%for ext in extensions:
   %if ext.has_features:
      <%helpers:guard ext="${ext}">
         %if ext.features_promoted:
               if (info->have_${ext.name_with_vendor()} && !info->have_vulkan${ext.core_since.struct()}) {
         %else:
               if (info->have_${ext.name_with_vendor()}) {
         %endif
                  info->${ext.field("feats")}.sType = ${ext.stype("FEATURES")};
                  info->${ext.field("feats")}.pNext = info->feats.pNext;
                  info->feats.pNext = &info->${ext.field("feats")};
               }
      </%helpers:guard>
   %endif
%endfor

   return true;

fail:
   return false;
}

void
zink_verify_device_extensions(struct zink_screen *screen)
{
%for ext in extensions:
   %if registry.in_registry(ext.name):
      <%helpers:guard ext="${ext}">
            if (screen->info.have_${ext.name_with_vendor()}) {
         %for cmd in registry.get_registry_entry(ext.name).device_commands:
            %if cmd.name().find("Win32") != -1:
            #ifdef _WIN32
            %endif
            %if ext.is_promoted_to_khr:
                  if (!screen->vk.${cmd.name()}) {
                     screen->vk.${cmd.name()} = (PFN_${cmd.full_name})screen->vk.${cmd.name().replace("EXT", "KHR")}; /* promoted from EXT */
                  }
            %endif
            %if ext.core_since:
               %if not cmd.not_promoted:
                     if (!screen->vk.${cmd.name().rstrip(ext.vendor())}) {
                        screen->vk.${cmd.name().rstrip(ext.vendor())} = screen->vk.${cmd.name()}; /* promoted to core */
                     }
               %else:
                     /* ${cmd.full_name} is not promoted */
               %endif
            %endif
                  if (!screen->vk.${cmd.name()}) {
            #ifndef NDEBUG
                     screen->vk.${cmd.name()} = (PFN_${cmd.full_name})zink_stub_${cmd.name()};
            #else
                     screen->vk.${cmd.name()} = (PFN_${cmd.full_name})zink_stub_function_not_loaded;
            #endif
                  }
            %if cmd.name().find("Win32") != -1:
            #endif
            %endif
         %endfor
            }
         %if ext.core_since:
               else if (screen->info.have_vulkan${ext.core_since.struct()}) {
            %for cmd in registry.get_registry_entry(ext.name).device_commands:
               %if cmd.name().find("Win32") != -1:
               #ifdef _WIN32
               %endif
               %if not cmd.not_promoted:
                     screen->vk.${cmd.name()} = (PFN_${cmd.full_name})screen->vk.${cmd.name().rstrip(ext.vendor())};
               %else:
                     /* ${cmd.full_name} is not promoted */
               %endif
               %if cmd.name().find("Win32") != -1:
               #endif
               %endif
            %endfor
               }
         %endif
      </%helpers:guard>
   %endif
%endfor
}

#ifndef NDEBUG
/* generated stub functions */
## remember the stub functions that are already generated
<% generated_funcs = set() %>

%for ext in extensions:
   %if registry.in_registry(ext.name):
      %for cmd in registry.get_registry_entry(ext.name).device_commands:
      ##
      ## some functions are added by multiple extensions, which creates duplication
      ## and thus redefinition of stubs (eg. vkCmdPushDescriptorSetWithTemplateKHR)
      ##
      %if cmd.name() in generated_funcs:
         <% continue %>
      %else:
         <% generated_funcs.add(cmd.name()) %>
      %endif
      void VKAPI_PTR
      zink_stub_${cmd.name()}()
      {
         mesa_loge("ZINK: ${cmd.full_name} is not loaded properly!");
         abort();
      }
      %endfor
   %endif
%endfor
#endif
"""


def replace_code(code: str, replacement: dict):
    for (k, v) in replacement.items():
        code = code.replace(k, v)
    
    return code


if __name__ == "__main__":
    try:
        header_path = sys.argv[1]
        impl_path = sys.argv[2]
        vkxml_path = sys.argv[3]

        header_path = path.abspath(header_path)
        impl_path = path.abspath(impl_path)
        vkxml_path = path.abspath(vkxml_path)
    except:
        print("usage: %s <path to .h> <path to .c> <path to vk.xml>" % sys.argv[0])
        exit(1)

    registry = ExtensionRegistry(vkxml_path)

    extensions = EXTENSIONS
    versions = VERSIONS
    replacement = REPLACEMENTS

    # Perform extension validation and set core_since for the extension if available
    error_count = 0
    for ext in extensions:
        if not registry.in_registry(ext.name):
            # disable validation for nonstandard extensions
            if ext.is_nonstandard:
                continue

            error_count += 1
            print("The extension {} is not registered in vk.xml - a typo?".format(ext.name))
            continue

        entry = registry.get_registry_entry(ext.name)

        if entry.ext_type != "device":
            error_count += 1
            print("The extension {} is {} extension - expected a device extension.".format(ext.name, entry.ext_type))
            continue

        ext.is_promoted_to_khr = entry.is_promoted_to_khr

        if ext.is_promoted_to_khr:
            latest_entry = registry.get_registry_entry(ext.name.replace("VK_EXT_", "VK_KHR_"))
        else:
            latest_entry = entry

        if ext.has_features:
            if not (entry.features_struct and ext.physical_device_struct("Features") == entry.features_struct):
                error_count += 1
                print("The extension {} does not provide a features struct.".format(ext.name))
            ext.features_promoted = latest_entry.features_promoted

        if ext.has_properties:
            if not (entry.properties_struct and ext.physical_device_struct("Properties") == entry.properties_struct):
                error_count += 1
                print("The extension {} does not provide a properties struct.".format(ext.name))
            ext.properties_promoted = latest_entry.properties_promoted
            ext.needs_double_load = latest_entry.needs_double_load

        if latest_entry.promoted_in and latest_entry.promoted_in <= versions[-1].struct_version:
            ext.core_since = Version((*latest_entry.promoted_in, 0))
        else:
            # even if the ext is promoted in a newer VK version, consider it
            # unpromoted until there's an entry for that VK version in VERSIONS
            ext.features_promoted = False
            ext.properties_promoted = False

    if error_count > 0:
        print("zink_device_info.py: Found {} error(s) in total. Quitting.".format(error_count))
        exit(1)

    lookup = TemplateLookup()
    lookup.put_string("helpers", include_template)

    with open(header_path, "w", encoding='utf-8') as header_file:
        header = Template(header_code, lookup=lookup).render(extensions=extensions, versions=versions, registry=registry).strip()
        header = replace_code(header, replacement)
        print(header, file=header_file)

    with open(impl_path, "w", encoding='utf-8') as impl_file:
        impl = Template(impl_code, lookup=lookup).render(extensions=extensions, versions=versions, registry=registry).strip()
        impl = replace_code(impl, replacement)
        print(impl, file=impl_file)



#
# Copyright © 2021 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import argparse
import sys

# List of the default tracepoints enabled. By default most tracepoints are
# enabled, set tp_default=False to disable them by default.
#
# Currently only stall is disabled by default
intel_default_tps = []

#
# Tracepoint definitions:
#
def define_tracepoints(args):
    from u_trace import Header, HeaderScope
    from u_trace import ForwardDecl
    from u_trace import Tracepoint
    from u_trace import TracepointArg as Arg
    from u_trace import TracepointArgStruct as ArgStruct

    Header('intel_driver_ds.h', scope=HeaderScope.SOURCE)
    Header('vulkan/vulkan_core.h', scope=HeaderScope.SOURCE|HeaderScope.PERFETTO)
    Header('blorp/blorp_priv.h', scope=HeaderScope.HEADER)
    Header('ds/intel_driver_ds.h', scope=HeaderScope.HEADER)

    def begin_end_tp(name, tp_args=[], tp_struct=None, tp_print=None,
                     tp_default_enabled=True, end_pipelined=True,
                     compute=False, repeat_last=False,
                     need_cs_param=False,
                     toggle_name=None):
        global intel_default_tps
        if toggle_name is None:
            toggle_name = name
        if tp_default_enabled and toggle_name not in intel_default_tps:
            intel_default_tps.append(name)

        # Preprocess arguments to handle display_as_hex
        processed_args = []
        for arg in tp_args:
            # Manually handle display_as_hex by modifying the format
            if "hash" in arg.var:
                arg = Arg(type=arg.type, var=arg.var, c_format='%#x')  # Convert to hex format
            processed_args.append(arg)

        Tracepoint('intel_begin_{0}'.format(name),
                   toggle_name=toggle_name,
                   tp_perfetto='intel_ds_begin_{0}'.format(name),
                   need_cs_param=need_cs_param)
        tp_flags = []
        if end_pipelined:
            if compute:
                tp_flags.append('INTEL_DS_TRACEPOINT_FLAG_END_CS')
            elif repeat_last:
                tp_flags.append('INTEL_DS_TRACEPOINT_FLAG_REPEAST_LAST')
            else:
                tp_flags.append('INTEL_DS_TRACEPOINT_FLAG_END_OF_PIPE')
        Tracepoint('intel_end_{0}'.format(name),
                   toggle_name=toggle_name,
                   args=processed_args,
                   tp_struct=tp_struct,
                   tp_perfetto='intel_ds_end_{0}'.format(name),
                   tp_print=tp_print,
                   tp_flags=tp_flags,
                   need_cs_param=need_cs_param)

    def draw_tp(name, tp_args=[]):
        begin_end_tp(name, tp_args=tp_args, toggle_name='draw',
                     tp_default_enabled=not args.android)

    # Frame tracepoints
    begin_end_tp('frame',
                 tp_args=[Arg(type='uint32_t', var='frame', c_format='%u'),],
                 end_pipelined=False,
                 need_cs_param=True)

    # Annotations for Queue(Begin|End)DebugUtilsLabelEXT
    begin_end_tp('queue_annotation',
                 tp_args=[Arg(type='unsigned', var='len'),
                          Arg(type='str', var='str', c_format='%s', length_arg='len + 1', copy_func='strncpy'),],
                 tp_struct=[Arg(type='uint8_t', name='dummy', var='0')],
                 end_pipelined=False,
                 need_cs_param=True)

    # Batch buffer tracepoints, only for Iris
    begin_end_tp('batch',
                 tp_args=[Arg(type='uint8_t', var='name', c_format='%hhu'),],
                 end_pipelined=False)

    # Command buffer tracepoints, only for Anv
    begin_end_tp('cmd_buffer',
                 tp_args=[Arg(type='uint64_t', var='command_buffer_handle', c_format='%" PRIu64 "',  perfetto_field=True),
                          Arg(type='uint8_t', var='level', c_format='%hhu'),],
                 end_pipelined=False)

    # Annotations for Cmd(Begin|End)DebugUtilsLabelEXT
    begin_end_tp('cmd_buffer_annotation',
                 tp_args=[Arg(type='uint64_t', var='command_buffer_handle', c_format='%" PRIu64 "',  perfetto_field=True),
                          Arg(type='unsigned', var='len'),
                          Arg(type='str', var='str', c_format='%s', length_arg='len + 1', copy_func='strncpy'),],
                 tp_struct=[Arg(type='uint8_t', name='dummy', var='0'),],
                 end_pipelined=True)

    # Transform feedback, only for Anv
    begin_end_tp('xfb',
                 end_pipelined=False)

    # STATE_BASE_ADDRESS emission, only for Anv
    begin_end_tp('sba',
                 tp_args=[Arg(type='uint8_t', var='mode', c_format='%hhu'),],
                 end_pipelined=False)

    # Dynamic rendering tracepoints, only for Anv
    begin_end_tp('render_pass',
                 tp_args=[Arg(type='uint64_t', var='command_buffer_handle', c_format='%" PRIu64 "',  perfetto_field=True),
                          Arg(type='uint16_t', var='width', c_format='%hu'),
                          Arg(type='uint16_t', var='height', c_format='%hu'),
                          Arg(type='uint8_t', var='att_count', c_format='%hhu'),
                          Arg(type='uint8_t', var='msaa', c_format='%hhu'),])

    # Blorp operations, Anv & Iris
    begin_end_tp('blorp',
                 tp_args=[Arg(type='enum blorp_op', var='op', c_format='%s', to_prim_type='blorp_op_to_name({})'),
                          Arg(type='uint32_t', var='width', c_format='%u'),
                          Arg(type='uint32_t', var='height', c_format='%u'),
                          Arg(type='uint32_t', var='samples', c_format='%u'),
                          Arg(type='enum blorp_shader_pipeline', var='shader_pipe', c_format='%s', to_prim_type='blorp_shader_pipeline_to_name({})'),
                          Arg(type='enum isl_format', var='dst_fmt', c_format='%s', to_prim_type='isl_format_get_short_name({})'),
                          Arg(type='enum isl_format', var='src_fmt', c_format='%s', to_prim_type='isl_format_get_short_name({})'),
                          Arg(type='uint8_t', var='predicated', c_format='%hhu'),
                          ],
                 tp_default_enabled=not args.android)

    # vkCmdWriteBufferMarker*, only for Anv
    begin_end_tp('write_buffer_marker',
                 end_pipelined=False)

    # Indirect draw generation, only for Anv
    begin_end_tp('generate_draws')
    begin_end_tp('generate_commands')

    # vkCmdResetQuery, only for Anv
    begin_end_tp('query_clear_blorp',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u')])
    begin_end_tp('query_clear_cs',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u')],
                 end_pipelined=False)

    # vkCmdCopyQueryResults, only for Anv
    begin_end_tp('query_copy_cs',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u')],
                 end_pipelined=False)
    begin_end_tp('query_copy_shader',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u')])

    # Various draws/dispatch, Anv & Iris
    draw_tp('draw',
            tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])
    draw_tp('draw_multi',
            tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])
    draw_tp('draw_indexed',
            tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])
    draw_tp('draw_indexed_multi',
            tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])
    draw_tp('draw_indirect_byte_count',
            tp_args=[Arg(type='uint32_t', var='instance_count', c_format='%u'),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])
    draw_tp('draw_indirect',
            tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u'),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])
    draw_tp('draw_indexed_indirect',
            tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u'),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])
    draw_tp('draw_indirect_count',
            tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u',
                         is_indirect=True),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])
    draw_tp('draw_indexed_indirect_count',
            tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u',
                         is_indirect=True),
                     Arg(type='uint32_t', var='vs_hash', c_format='%u'),
                     Arg(type='uint32_t', var='fs_hash', c_format='%u')])

    draw_tp('draw_mesh',
            tp_args=[Arg(type='uint32_t', var='group_x', c_format='%u'),
                     Arg(type='uint32_t', var='group_y', c_format='%u'),
                     Arg(type='uint32_t', var='group_z', c_format='%u'),])
    draw_tp('draw_mesh_indirect',
            tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u'),])
    draw_tp('draw_mesh_indirect_count',
            tp_args=[Arg(type='uint32_t', var='draw_count', c_format='%u',
                              is_indirect=True),])

    begin_end_tp('compute',
                 tp_args=[Arg(type='uint32_t', var='group_x', c_format='%u'),
                          Arg(type='uint32_t', var='group_y', c_format='%u'),
                          Arg(type='uint32_t', var='group_z', c_format='%u'),
                          Arg(type='uint32_t', var='cs_hash', c_format='%u')],
                 compute=True)

    begin_end_tp('compute_indirect',
                 tp_args=[ArgStruct(type='VkDispatchIndirectCommand', var='size',
                          is_indirect=True, c_format="%ux%ux%u",
                          fields=['x', 'y', 'z']),
                          Arg(type='uint32_t', var='cs_hash', c_format='%u')],
                 compute=True)

    # Used to identify copies generated by utrace
    begin_end_tp('trace_copy',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),])
    begin_end_tp('trace_copy_cb',
                 tp_args=[Arg(type='uint32_t', var='count', c_format='%u'),],
                 need_cs_param=True)

    begin_end_tp('as_build',
                 tp_args=[Arg(type='uint32_t', var='n_tlas', c_format='%u'),
                          Arg(type='uint32_t', var='n_blas', c_format='%u'),])
    begin_end_tp('as_build_leaves', repeat_last=True)
    begin_end_tp('as_morton_generate', repeat_last=True)
    begin_end_tp('as_morton_sort', repeat_last=True)
    begin_end_tp('as_lbvh_build_internal', repeat_last=True)
    begin_end_tp('as_ploc_build_internal', repeat_last=True)
    begin_end_tp('as_encode', repeat_last=True,
                  tp_args=[Arg(type='uint32_t', var='pass', c_format='%u'),
                           Arg(type='uint32_t', var='key', c_format='%x'),
                           Arg(type='uint32_t', var='n_leaves', c_format='%u'),
                           Arg(type='uint32_t', var='n_ir_leaves', c_format='%u')])
    begin_end_tp('as_copy', repeat_last=True)

    begin_end_tp('rays',
                 tp_args=[Arg(type='uint32_t', var='group_x', c_format='%u'),
                          Arg(type='uint32_t', var='group_y', c_format='%u'),
                          Arg(type='uint32_t', var='group_z', c_format='%u'),],
                 compute=True)

    def flag_bits(args):
        bits = [Arg(type='enum intel_ds_stall_flag', name='flags', var='decode_cb(flags)', c_format='0x%x')]
        for a in args:
            bits.append(Arg(type='bool', name=a[1], var='__entry->flags & INTEL_DS_{0}_BIT'.format(a[0]), c_format='%u'))
        return bits

    def stall_args(stall_bits):
        fmt = ''
        exprs = []
        for a in stall_bits:
            fmt += '%s'
            exprs.append('(__entry->flags & INTEL_DS_{0}_BIT) ? "+{1}" : ""'.format(a[0], a[1]))
        fmt += ' : %s%s%s%s%s%s%s'
        exprs.append('(__entry->reason1) ? __entry->reason1 : "unknown"')
        exprs.append('(__entry->reason2) ? "; " : ""')
        exprs.append('(__entry->reason2) ? __entry->reason2 : ""')
        exprs.append('(__entry->reason3) ? "; " : ""')
        exprs.append('(__entry->reason3) ? __entry->reason3 : ""')
        exprs.append('(__entry->reason4) ? "; " : ""')
        exprs.append('(__entry->reason4) ? __entry->reason4 : ""')
        # To printout flags
        # fmt += '(0x%08x)'
        # exprs.append('__entry->flags')
        fmt = [fmt]
        fmt += exprs
        return fmt

    def barrier_args(stage_bits, stall_bits):
        fmt = ''
        exprs = []

        fmt += '%s '
        exprs.append('(__entry->type == INTEL_DS_BARRIER_TYPE_IMMEDIATE) ? "IMMEDIATE" : ' +
                     '(__entry->type == INTEL_DS_BARRIER_TYPE_SIGNAL)    ? "SIGNAL" : ' +
                     '(__entry->type == INTEL_DS_BARRIER_TYPE_WAIT)      ? "WAIT" : "unknown"')

        for a in stages_bits:
            fmt += '%s'
            exprs.append('(__entry->signal_stages & INTEL_DS_STAGES_{0}_BIT) ? "+{1}" : ""'.format(a[0], a[1]))
        fmt += '->'
        for a in stages_bits:
            fmt += '%s'
            exprs.append('(__entry->wait_stages & INTEL_DS_STAGES_{0}_BIT) ? "+{1}" : ""'.format(a[0], a[1]))
        fmt += ': '

        for a in stall_bits:
            fmt += '%s'
            exprs.append('(__entry->flags & INTEL_DS_{0}_BIT) ? "+{1}" : ""'.format(a[0], a[1]))
        fmt += ' : %s%s%s%s%s%s%s'
        exprs.append('(__entry->reason1) ? __entry->reason1 : "unknown"')
        exprs.append('(__entry->reason2) ? "; " : ""')
        exprs.append('(__entry->reason2) ? __entry->reason2 : ""')
        exprs.append('(__entry->reason3) ? "; " : ""')
        exprs.append('(__entry->reason3) ? __entry->reason3 : ""')
        exprs.append('(__entry->reason4) ? "; " : ""')
        exprs.append('(__entry->reason4) ? __entry->reason4 : ""')
        # To printout flags
        # fmt += '(0x%08x)'
        # exprs.append('__entry->flags')
        fmt = [fmt]
        fmt += exprs
        return fmt

    stall_flags = [['DEPTH_CACHE_FLUSH',             'depth_flush'],
                   ['DATA_CACHE_FLUSH',              'dc_flush'],
                   ['HDC_PIPELINE_FLUSH',            'hdc_flush'],
                   ['RENDER_TARGET_CACHE_FLUSH',     'rt_flush'],
                   ['TILE_CACHE_FLUSH',              'tile_flush'],
                   ['STATE_CACHE_INVALIDATE',        'state_inval'],
                   ['CONST_CACHE_INVALIDATE',        'const_inval'],
                   ['VF_CACHE_INVALIDATE',           'vf_inval'],
                   ['TEXTURE_CACHE_INVALIDATE',      'tex_inval'],
                   ['INST_CACHE_INVALIDATE',         'ic_inval'],
                   ['STALL_AT_SCOREBOARD',           'pb_stall'],
                   ['DEPTH_STALL',                   'depth_stall'],
                   ['CS_STALL',                      'cs_stall'],
                   ['UNTYPED_DATAPORT_CACHE_FLUSH',  'udp_flush'],
                   ['PSS_STALL_SYNC',                'pss_stall'],
                   ['END_OF_PIPE',                   'eop'],
                   ['CCS_CACHE_FLUSH',               'ccs_flush']]

    stages_bits = [['TOP',    'top'],
                   ['GEOM',   'geom'],
                   ['RASTER', 'raster'],
                   ['DEPTH',  'depth'],
                   ['PIXEL',  'pixel'],
                   ['COLOR',  'color'],
                   ['GPGPU',  'gpgpu']]

    begin_end_tp('barrier',
                 tp_args=[ArgStruct(type='uint8_t', var='type'),
                          ArgStruct(type='uint32_t', var='signal_stages'),
                          ArgStruct(type='uint32_t', var='wait_stages'),
                          ArgStruct(type='intel_ds_stages_cb_t', var='decode_stage_cb'),
                          ArgStruct(type='uint32_t', var='flags'),
                          ArgStruct(type='intel_ds_stall_cb_t', var='decode_flags_cb'),
                          ArgStruct(type='const char *', var='reason1'),
                          ArgStruct(type='const char *', var='reason2'),
                          ArgStruct(type='const char *', var='reason3'),
                          ArgStruct(type='const char *', var='reason4'),],
                 tp_struct=[Arg(type='uint8_t', name='type', var='type', c_format='0x%hhx'),
                            Arg(type='uint8_t', name='signal_stages', var='decode_stage_cb(signal_stages)', c_format='0x%hhx'),
                            Arg(type='uint8_t', name='wait_stages', var='decode_stage_cb(wait_stages)', c_format='0x%hhx'),
                            Arg(type='uint32_t', name='flags', var='decode_flags_cb(flags)', c_format='0x%x'),
                            Arg(type='const char *', name='reason1', var='reason1', c_format='%s'),
                            Arg(type='const char *', name='reason2', var='reason2', c_format='%s'),
                            Arg(type='const char *', name='reason3', var='reason3', c_format='%s'),
                            Arg(type='const char *', name='reason4', var='reason4', c_format='%s'),],
                 tp_print=barrier_args(stages_bits, stall_flags),
                 tp_default_enabled=False,
                 end_pipelined=False)

    begin_end_tp('stall',
                 tp_args=[ArgStruct(type='uint32_t', var='flags'),
                          ArgStruct(type='intel_ds_stall_cb_t', var='decode_cb'),
                          ArgStruct(type='const char *', var='reason1'),
                          ArgStruct(type='const char *', var='reason2'),
                          ArgStruct(type='const char *', var='reason3'),
                          ArgStruct(type='const char *', var='reason4'),],
                 tp_struct=[Arg(type='uint32_t', name='flags', var='decode_cb(flags)', c_format='0x%x'),
                            Arg(type='const char *', name='reason1', var='reason1', c_format='%s'),
                            Arg(type='const char *', name='reason2', var='reason2', c_format='%s'),
                            Arg(type='const char *', name='reason3', var='reason3', c_format='%s'),
                            Arg(type='const char *', name='reason4', var='reason4', c_format='%s'),],
                 tp_print=stall_args(stall_flags),
                 tp_default_enabled=False,
                 end_pipelined=False)


def generate_code(args):
    from u_trace import utrace_generate
    from u_trace import utrace_generate_perfetto_utils

    utrace_generate(cpath=args.utrace_src, hpath=args.utrace_hdr,
                    ctx_param='struct intel_ds_device *dev',
                    trace_toggle_name='intel_gpu_tracepoint',
                    trace_toggle_defaults=intel_default_tps)
    utrace_generate_perfetto_utils(hpath=args.perfetto_hdr,
                                   basename="intel_tracepoint")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--import-path', required=True)
    parser.add_argument('--utrace-src', required=True)
    parser.add_argument('--utrace-hdr', required=True)
    parser.add_argument('--perfetto-hdr', required=True)
    parser.add_argument('--android', action="store_true")
    args = parser.parse_args()
    sys.path.insert(0, args.import_path)
    define_tracepoints(args)
    generate_code(args)


if __name__ == '__main__':
    main()



# Copyright © 2024 Imagination Technologies Ltd.
# SPDX-License-Identifier: MIT

from mako.template import Template, exceptions
from pco_ops import *

template = """/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * SPDX-License-Identifier: MIT
 */

#ifndef PCO_OPS_H
#define PCO_OPS_H

/**
 * \\file pco_ops.h
 *
 * \\brief PCO op definitions and functions.
 */

#include "util/macros.h"

#include <assert.h>
#include <stdbool.h>
#include <stdint.h>

#define _PCO_OP_MAX_DESTS ${max([op.num_dests for op in ops.values()])}U
#define _PCO_OP_MAX_SRCS ${max([op.num_srcs for op in ops.values()])}U
#define _PCO_OP_MAX_MODS ${max([len(op.op_mods) for op in ops.values()])}U

/** Ops. */
#define _PCO_OP_COUNT ${len(ops) + 1}U
enum pco_op {
   PCO_OP_NONE,
% for op in ops.values():
   ${op.cname.upper()},
% endfor
};

/** Op mods. */
#define _PCO_OP_MOD_COUNT ${len(op_mods) + 1}U
enum pco_op_mod {
   PCO_OP_MOD_NONE,
% for op_mod in op_mods.values():
   ${op_mod.cname},
% endfor
};

/** Ref mods. */
#define _PCO_REF_MOD_COUNT ${len(ref_mods) + 1}U
enum pco_ref_mod {
   PCO_REF_MOD_NONE,
% for ref_mod in ref_mods.values():
   ${ref_mod.cname},
% endfor
};
#endif /* PCO_OPS_H */"""

def main():
   try:
      print(Template(template).render(ops=ops, op_mods=op_mods, ref_mods=ref_mods))
   except:
       raise Exception(exceptions.text_error_template().render())

if __name__ == '__main__':
   main()



#
# Copyright 2017-2019 Advanced Micro Devices, Inc.
#
# SPDX-License-Identifier: MIT
#
"""
Helper script to merge register database JSON files.

  usage: python3 mergedbs.py [FILES...]

Will merge the given JSON files and output the result on stdout.
"""

from collections import defaultdict
import json
import re
import sys

from regdb import RegisterDatabase, deduplicate_enums, deduplicate_register_types

def main():
    regdb = RegisterDatabase()
    for filename in sys.argv[1:]:
        with open(filename, 'r') as filp:
            regdb.update(RegisterDatabase.from_json(json.load(filp)))

    deduplicate_enums(regdb)
    deduplicate_register_types(regdb)

    print(regdb.encode_json_pretty())


if __name__ == '__main__':
    main()

# kate: space-indent on; indent-width 4; replace-tabs on;



#!/usr/bin/env python3

import argparse
import git
from git import Repo
import glob
import os
import shutil
import sys

# where should files go
dirs = {
  '/classes/3d/': 'classes/',
  '/classes/compute/': 'classes/',
  '/classes/dma-copy/': 'classes/',
  '/classes/host/': 'classes/',
  '/classes/memory-to-memory-format/': 'classes/',
  '/classes/inline-to-memory/': 'classes/',
  '/classes/twod/': 'classes/',
}
branch = 'master'
target = os.path.abspath(os.path.dirname(__file__)) + "/nvidia/"

parser = argparse.ArgumentParser(description='Updates Nvidia header files from git.')
parser.add_argument('git_path', type=str, help='Path to the open-gpu-doc repo')

args = parser.parse_args()
repo_path = os.path.abspath(args.git_path)

# 1. create repo object
try:
    repo = Repo(repo_path)
    assert not repo.bare
except git.exc.NoSuchPathError:
    print("{} doesn't point to a git repository".format(repo_path))
    sys.exit(-1)

# 2. update repo
repo.remotes.origin.fetch()
repo.git.checkout(branch)
repo.git.rebase('origin/' + branch)

# 3. check if all needed directories exist
for dir in dirs.keys():
    path = repo_path + dir
    if not os.path.isdir(path):
        print(dir + " does not exist in repository. Was the correct repository choosen?")
        sys.exit(-1)

# 4. copy over files
for src, dest in dirs.items():
    src = repo_path + src
    dest = target + dest
    for header in glob.glob(src + "*.h"):
        print(header + " => " + dest)
        shutil.copy(header, dest)




# (C) Copyright IBM Corporation 2004
# All Rights Reserved.
# Copyright (c) 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# on the rights to use, copy, modify, merge, publish, distribute, sub
# license, and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
# IBM AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# Authors:
#    Ian Romanick <idr@us.ibm.com>

import argparse

import gl_XML
import static_data
import license


class PrintGlTable(gl_XML.gl_print_base):
    def __init__(self):
        gl_XML.gl_print_base.__init__(self)

        self.header_tag = '_GLAPI_TABLE_H_'
        self.name = "gl_table.py (from Mesa)"
        self.license = license.bsd_license_template % ( \
"""Copyright (C) 1999-2003  Brian Paul   All Rights Reserved.
(C) Copyright IBM Corporation 2004""", "BRIAN PAUL, IBM")
        return

    def printBody(self, api):
        for f in api.functionIterateByOffset():
            arg_string = f.get_parameter_string()
            print('   %s (GLAPIENTRYP %s)(%s); /* %d */' % (
                f.return_type, f.name, arg_string, f.offset))

    def printRealHeader(self):
        print('#include "util/glheader.h"')
        print('')
        print('#ifdef __cplusplus')
        print('extern "C" {')
        print('#endif')
        print('')
        print('#if defined(_WIN32) && defined(_WINDOWS_)')
        print('#error "Should not include <windows.h> here"')
        print('#endif')
        print('')
        print('struct _glapi_table')
        print('{')
        return

    def printRealFooter(self):
        print('};')
        print('')
        print('#ifdef __cplusplus')
        print('}')
        print('#endif')
        return


class PrintRemapTable(gl_XML.gl_print_base):
    def __init__(self):
        gl_XML.gl_print_base.__init__(self)

        self.header_tag = '_DISPATCH_H_'
        self.name = "gl_table.py (from Mesa)"
        self.license = license.bsd_license_template % (
            "(C) Copyright IBM Corporation 2005", "IBM")
        return


    def printRealHeader(self):
        print("""
/**
 * \\file main/dispatch.h
 * Macros for handling GL dispatch tables.
 *
 * For each known GL function, there are 3 macros in this file.  The first
 * macro is named CALL_FuncName and is used to call that GL function using
 * the specified dispatch table.  The other 2 macros, called GET_FuncName
 * can SET_FuncName, are used to get and set the dispatch pointer for the
 * named function in the specified dispatch table.
 */

#include "util/glheader.h"
""")
        return


    def printBody(self, api):
        print('#define CALL_by_offset(disp, cast, offset, parameters) \\')
        print('    (*(cast (GET_by_offset(disp, offset)))) parameters')
        print('#define GET_by_offset(disp, offset) \\')
        print('    (offset >= 0) ? (((_glapi_proc *)(disp))[offset]) : NULL')
        print('#define SET_by_offset(disp, offset, fn) \\')
        print('    do { \\')
        print('        assert(offset >= 0 && offset < _gloffset_COUNT); \\')
        print('        ( (_glapi_proc *) (disp) )[offset] = (_glapi_proc) fn; \\')
        print('    } while(0)')
        print('')

        abi_functions = [f for f in api.functionIterateByOffset()]

        print('/* total number of offsets below */')
        print('#define _gloffset_COUNT %d' % (static_data.function_count))
        print('')

        for f in abi_functions:
            print('#define _gloffset_%s %d' % (f.name, f.offset))

        print('')

        for f in abi_functions:
            arg_string = gl_XML.create_parameter_string(f.parameters, 0)

            print('typedef %s (GLAPIENTRYP _glptr_%s)(%s);' % (f.return_type, f.name, arg_string))
            print('#define CALL_{0}(disp, parameters) (* GET_{0}(disp)) parameters'.format(f.name))
            print('#define GET_{0}(disp) ((_glptr_{0})(GET_by_offset((disp), _gloffset_{0})))'.format(f.name))
            print("""#define SET_{0}(disp, func) do {{ \\
   {1} (GLAPIENTRYP fn)({2}) = func; \\
   SET_by_offset(disp, _gloffset_{0}, fn); \\
}} while (0)
""".format(f.name, f.return_type, arg_string))

        return


def _parser():
    """Parse arguments and return a namespace."""
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filename',
                        default='gl_API.xml',
                        metavar="input_file_name",
                        dest='file_name',
                        help="Path to an XML description of OpenGL API.")
    parser.add_argument('-m', '--mode',
                        choices=['table', 'dispatch'],
                        default='table',
                        metavar="mode",
                        help="Generate either a table or a dispatch")
    return parser.parse_args()


def main():
    """Main function."""
    args = _parser()

    api = gl_XML.parse_GL_API(args.file_name)

    if args.mode == "table":
        printer = PrintGlTable()
    elif args.mode == "dispatch":
        printer = PrintRemapTable()

    printer.Print(api)


if __name__ == '__main__':
    main()



# Copyright © 2023 Igalia S.L.
# SPDX-License-Identifier: MIT

"""Create shortcuts for casting Vulkan structs when knowing their stype."""

import argparse
import functools
import os
import re
import textwrap
import xml.etree.ElementTree as et

from mako.template import Template
from vk_extensions import Extension, filter_api, get_all_required

COPYRIGHT = textwrap.dedent(u"""\
     * Copyright © 2023 Igalia S.L.
     * SPDX-License-Identifier: MIT
     """)

H_TEMPLATE = Template(textwrap.dedent(u"""\
    /* Autogenerated file -- do not edit
     * generated by ${file}
     *
     ${copyright}
     */

    #ifndef MESA_VK_STRUCT_CASTS_H
    #define MESA_VK_STRUCT_CASTS_H

    #include <vulkan/vulkan.h>

    #ifdef __cplusplus
    extern "C" {
    #endif

    % for s in structs:
    #define ${s.stype}_cast ${s.name}
    % endfor

    #ifdef __cplusplus
    } /* extern "C" */
    #endif

    #endif
    """))


class VkStruct(object):
    """Simple struct-like class representing a single Vulkan struct identified with a VkStructureType"""
    def __init__(self, name, stype):
        self.name = name
        self.stype = stype


def struct_get_stype(xml_node):
    for member in xml_node.findall('./member'):
        name = member.findall('./name')
        if len(name) > 0 and name[0].text == "sType":
            return member.get('values')
    return None


def parse_xml(filename, structs, beta):
    xml = et.parse(filename)
    api = 'vulkan'

    required_types = get_all_required(xml, 'type', api, beta)

    for struct_type in xml.findall('./types/type[@category="struct"]'):
        if not filter_api(struct_type, api):
            continue

        name = struct_type.attrib['name']
        if name not in required_types:
            continue

        stype = struct_get_stype(struct_type)
        if stype is not None:
            structs.append(VkStruct(name, stype))

    for struct_type in xml.findall('.//enum[@alias][@extends=\'VkStructureType\']'):
        name = struct_type.attrib['name']
        alias = struct_type.attrib['alias']
        structs.append(VkStruct(alias + "_cast", name))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--xml', required=True,
                        help='Vulkan API XML files',
                        action='append',
                        dest='xml_files')
    parser.add_argument('--out',
                        help='Output H file',
                        required=True)
    parser.add_argument('--beta', required=True, help='Enable beta extensions.')

    args = parser.parse_args()

    structs = []

    for filename in args.xml_files:
        parse_xml(filename, structs, args.beta)

    structs = sorted(structs, key=lambda s: s.name)

    for template, file_ in [(H_TEMPLATE, args.out)]:
        with open(file_, 'w', encoding='utf-8') as f:
            f.write(template.render(
                file=os.path.basename(__file__),
                structs=structs,
                copyright=COPYRIGHT))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
##########################################################################
# 
# Copyright 2008 VMware, Inc.
# All Rights Reserved.
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sub license, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice (including the
# next paragraph) shall be included in all copies or substantial portions
# of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
# IN NO EVENT SHALL VMWARE AND/OR ITS SUPPLIERS BE LIABLE FOR
# ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 
##########################################################################


from parse import *


if __name__ == '__main__':
    Main().main()



#encoding=utf-8

import argparse
import ast
import intel_genxml
import re
import sys
import copy
import textwrap
from util import *

license =  """/*
 * Copyright (C) 2016 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */
"""

pack_header = """%(license)s

/* Instructions, enums and structures for %(platform)s.
 *
 * This file has been generated, do not hand edit.
 */

#ifndef %(guard)s
#define %(guard)s

#ifndef __OPENCL_VERSION__
#include <stdio.h>
#include "util/bitpack_helpers.h"
#include "genX_helpers.h"
#else
#include "genX_cl_helpers.h"
#endif

"""

def num_from_str(num_str):
    if num_str.lower().startswith('0x'):
        return int(num_str, base=16)

    assert not num_str.startswith('0'), 'octals numbers not allowed'
    return int(num_str)

def bool_from_str(bool_str):
    options = { "true": True, "false": False }
    return options[bool_str];

class Field(object):
    ufixed_pattern = re.compile(r"u(\d+)\.(\d+)")
    sfixed_pattern = re.compile(r"s(\d+)\.(\d+)")

    def __init__(self, parser, attrs):
        self.parser = parser
        if "name" in attrs:
            self.name = safe_name(attrs["name"])

        dword = int(attrs["dword"])
        end_bit, start_bit = map(int, attrs["bits"].split(":"))

        self.start = dword * 32 + start_bit
        self.end = dword * 32 + end_bit

        self.type = attrs["type"]
        self.nonzero = bool_from_str(attrs.get("nonzero", "false"))
        self.prefix = attrs["prefix"] if "prefix" in attrs else None

        assert self.start <= self.end, \
               'field {} has end ({}) < start ({})'.format(self.name, self.end,
                                                           self.start)
        if self.type == 'bool':
            assert self.end == self.start, \
                   'bool field ({}) is too wide'.format(self.name)

        if "default" in attrs:
            # Base 0 recognizes 0x, 0o, 0b prefixes in addition to decimal ints.
            self.default = int(attrs["default"], base=0)
        else:
            self.default = None

        ufixed_match = Field.ufixed_pattern.match(self.type)
        if ufixed_match:
            self.type = 'ufixed'
            self.fractional_size = int(ufixed_match.group(2))

        sfixed_match = Field.sfixed_pattern.match(self.type)
        if sfixed_match:
            self.type = 'sfixed'
            self.fractional_size = int(sfixed_match.group(2))

    def is_builtin_type(self):
        builtins =  [ 'address', 'bool', 'float', 'ufixed',
                      'offset', 'sfixed', 'offset', 'int', 'uint',
                      'mbo', 'mbz' ]
        return self.type in builtins

    def is_struct_type(self):
        return self.type in self.parser.structs

    def is_enum_type(self):
        return self.type in self.parser.enums

    def emit_template_struct(self, dim):
        if self.type == 'address':
            type = 'uint64_t' if self.parser.opencl else '__gen_address_type'
        elif self.type == 'bool':
            type = 'bool'
        elif self.type == 'float':
            type = 'float'
        elif self.type == 'ufixed':
            type = 'float'
        elif self.type == 'sfixed':
            type = 'float'
        elif self.type == 'uint' and self.end - self.start > 32:
            type = 'uint64_t'
        elif self.type == 'offset':
            type = 'uint64_t'
        elif self.type == 'int':
            type = 'int32_t'
        elif self.type == 'uint':
            type = 'uint32_t'
        elif self.is_struct_type():
            type = 'struct ' + self.parser.gen_prefix(safe_name(self.type))
        elif self.is_enum_type():
            type = 'enum ' + self.parser.gen_prefix(safe_name(self.type))
        elif self.type == 'mbo' or self.type == 'mbz':
            return
        else:
            print("#error unhandled type: %s" % self.type)
            return

        print("   %-36s %s%s;" % (type, self.name, dim))

        prefix = self.prefix + '_' if self.prefix else ''

        for value in self.values:
            name = value.name
            if self.prefix and value.name[0] == '_':
                name = name[1:]

            print("#define %-40s %d" % (prefix + name, value.value))

class Group(object):
    def __init__(self, parser, parent, start, count, size):
        self.parser = parser
        self.parent = parent
        self.start = start
        self.count = count
        self.size = size
        self.fields = []

    def emit_template_struct(self, dim):
        if self.count == 0:
            print("   /* variable length fields follow */")
        else:
            if self.count > 1:
                dim = "%s[%d]" % (dim, self.count)

            for field in self.fields:
                field.emit_template_struct(dim)

    class DWord:
        def __init__(self):
            self.size = 32
            self.fields = []
            self.address = None

    def collect_dwords(self, dwords, start, dim):
        for field in self.fields:
            if isinstance(field, Group):
                if field.count == 1:
                    field.collect_dwords(dwords, start + field.start, dim)
                else:
                    for i in range(field.count):
                        field.collect_dwords(dwords,
                                             start + field.start + i * field.size,
                                             "%s[%d]" % (dim, i))
                continue

            index = (start + field.start) // 32
            if not index in dwords:
                dwords[index] = self.DWord()

            clone = copy.copy(field)
            clone.start = clone.start + start
            clone.end = clone.end + start
            clone.dim = dim
            dwords[index].fields.append(clone)

            if field.type == "address":
                # assert dwords[index].address == None
                dwords[index].address = clone

            # Coalesce all the dwords covered by this field. The two cases we
            # handle are where multiple fields are in a 64 bit word (typically
            # and address and a few bits) or where a single struct field
            # completely covers multiple dwords.
            while index < (start + field.end) // 32:
                if index + 1 in dwords and not dwords[index] == dwords[index + 1]:
                    dwords[index].fields.extend(dwords[index + 1].fields)
                dwords[index].size = 64
                dwords[index + 1] = dwords[index]
                index = index + 1

    def collect_dwords_and_length(self, repack=False):
        dwords = {}
        self.collect_dwords(dwords, 0, "")

        # Determine number of dwords in this group. If we have a size, use
        # that, since that'll account for MBZ dwords at the end of a group
        # (like dword 8 on BDW+ 3DSTATE_HS). Otherwise, use the largest dword
        # index we've seen plus one.
        if self.size > 0:
            length = self.size // 32
        elif dwords:
            length = max(dwords.keys()) + 1
        else:
            length = 0

        return (dwords, length)

    def emit_pack_function(self, dwords, length, repack=False):
        for index in range(length):
            # Handle MBZ dwords
            if not index in dwords:
                print("")
                print("   dw[%d] = 0;" % index)
                continue

            # For 64 bit dwords, we aliased the two dword entries in the dword
            # dict it occupies. Now that we're emitting the pack function,
            # skip the duplicate entries.
            dw = dwords[index]
            if index > 0 and index - 1 in dwords and dw == dwords[index - 1]:
                continue

            # Special case: only one field and it's a struct at the beginning
            # of the dword. In this case we pack directly into the
            # destination. This is the only way we handle embedded structs
            # larger than 32 bits.
            if len(dw.fields) == 1:
                field = dw.fields[0]
                name = field.name + field.dim
                if field.is_struct_type() and field.start % 32 == 0:
                    print("")
                    if repack:
                        if self.parser.opencl:
                            print("   %s_repack(&dw[%d], &origin[%d], &values->%s);" %
                                  (self.parser.gen_prefix(safe_name(field.type)), index, index, name))
                        else:
                            print("   %s_pack(data, &dw[%d], &origin[%d], &values->%s);" %
                                  (self.parser.gen_prefix(safe_name(field.type)), index, index, name))
                    else:
                        if self.parser.opencl:
                            print("   %s_pack(&dw[%d], &values->%s);" %
                                  (self.parser.gen_prefix(safe_name(field.type)), index, name))
                        else:
                            print("   %s_pack(data, &dw[%d], &values->%s);" %
                                  (self.parser.gen_prefix(safe_name(field.type)), index, name))
                    continue

            # Pack any fields of struct type first so we have integer values
            # to the dword for those fields.
            field_index = 0
            for field in dw.fields:
                if isinstance(field, Field) and field.is_struct_type():
                    name = field.name + field.dim
                    print("")
                    print("   uint32_t v%d_%d;" % (index, field_index))
                    if repack:
                        if self.parser.opencl:
                            print("   %s_repack(&v%d_%d, &origin[%d], &values->%s);" %
                                  (self.parser.gen_prefix(safe_name(field.type)), index, field_index, index, name))
                        else:
                            print("   %s_repack(data, &v%d_%d, &origin[%d], &values->%s);" %
                                  (self.parser.gen_prefix(safe_name(field.type)), index, field_index, index, name))
                    else:
                        if self.parser.opencl:
                            print("   %s_pack(&v%d_%d, &values->%s);" %
                                  (self.parser.gen_prefix(safe_name(field.type)), index, field_index, name))
                        else:
                            print("   %s_pack(data, &v%d_%d, &values->%s);" %
                                  (self.parser.gen_prefix(safe_name(field.type)), index, field_index, name))
                    field_index = field_index + 1

            print("")
            dword_start = index * 32
            if dw.address == None:
                address_count = 0
            else:
                address_count = 1

            # Assert in dont_use values
            for field in dw.fields:
                for value in field.values:
                    if value.dont_use:
                        print("   assert(values->%s != %s);" %
                              (field.name, field.prefix + "_" + value.name))

            if dw.size == 32 and dw.address == None:
                v = None
                print("   dw[%d] =" % index)
            elif len(dw.fields) > address_count or repack:
                v = "v%d" % index
                print("   const uint%d_t %s =" % (dw.size, v))
            else:
                v = "0"

            field_index = 0
            non_address_fields = []

            if repack:
                non_address_fields.append("origin[%d]" % index)
                if dw.size > 32:
                    non_address_fields.append("((uint64_t)origin[%d] << 32)" % (index + 1))

            for field in dw.fields:
                if field.type != "mbo" and field.type != "mbz" and field.type != "repack":
                    name = field.name + field.dim

                nz = "_nonzero" if field.nonzero else ""

                if field.type == "repack":
                    non_address_fields.append("origin[%d]" % index)
                elif field.type == "mbo":
                    non_address_fields.append("util_bitpack_ones(%d, %d)" % \
                        (field.start - dword_start, field.end - dword_start))
                elif field.type == "mbz":
                    assert not field.nonzero
                elif field.type == "address":
                    pass
                elif field.type == "uint":
                    non_address_fields.append("util_bitpack_uint%s(values->%s, %d, %d)" % \
                        (nz, name, field.start - dword_start, field.end - dword_start))
                elif field.is_enum_type():
                    non_address_fields.append("util_bitpack_uint%s(values->%s, %d, %d)" % \
                        (nz, name, field.start - dword_start, field.end - dword_start))
                elif field.type == "int":
                    non_address_fields.append("util_bitpack_sint%s(values->%s, %d, %d)" % \
                        (nz, name, field.start - dword_start, field.end - dword_start))
                elif field.type == "bool":
                    non_address_fields.append("util_bitpack_uint%s(values->%s, %d, %d)" % \
                        (nz, name, field.start - dword_start, field.end - dword_start))
                elif field.type == "float":
                    non_address_fields.append("util_bitpack_float%s(values->%s)" % (nz, name))
                elif field.type == "offset":
                    non_address_fields.append("__gen_offset%s(values->%s, %d, %d)" % \
                        (nz, name, field.start - dword_start, field.end - dword_start))
                elif field.type == 'ufixed':
                    non_address_fields.append("util_bitpack_ufixed%s(values->%s, %d, %d, %d)" % \
                        (nz, name, field.start - dword_start, field.end - dword_start, field.fractional_size))
                elif field.type == 'sfixed':
                    non_address_fields.append("util_bitpack_sfixed%s(values->%s, %d, %d, %d)" % \
                        (nz, name, field.start - dword_start, field.end - dword_start, field.fractional_size))
                elif field.is_struct_type():
                    non_address_fields.append("util_bitpack_uint(v%d_%d, %d, %d)" % \
                        (index, field_index, field.start - dword_start, field.end - dword_start))
                    field_index = field_index + 1
                else:
                    non_address_fields.append("/* unhandled field %s, type %s */\n" % \
                                              (name, field.type))

            if non_address_fields:
                print(" |\n".join("      " + f for f in non_address_fields) + ";")

            if dw.size == 32:
                if dw.address:
                    if self.parser.opencl:
                        print("   dw[%d] = __gen_address(values->%s, %d, %d) | %s;" %
                              (index, dw.address.name + field.dim,
                               dw.address.start - dword_start, dw.address.end - dword_start, v))
                    else:
                        print("   dw[%d] = __gen_address(data, &dw[%d], values->%s, %s, %d, %d);" %
                              (index, index, dw.address.name + field.dim, v,
                               dw.address.start - dword_start, dw.address.end - dword_start))
                continue

            if dw.address:
                v_address = "v%d_address" % index
                if self.parser.opencl:
                    print("   const uint64_t %s =\n      __gen_address(values->%s, %d, %d) | %s;" %
                          (v_address, dw.address.name + field.dim,
                           dw.address.start - dword_start, dw.address.end - dword_start, v))
                else:
                    print("   const uint64_t %s =\n      __gen_address(data, &dw[%d], values->%s, %s, %d, %d);" %
                          (v_address, index, dw.address.name + field.dim, v,
                           dw.address.start - dword_start, dw.address.end - dword_start))
                if len(dw.fields) > address_count:
                    print("   dw[%d] = %s;" % (index, v_address))
                    print("   dw[%d] = (%s >> 32) | (%s >> 32);" % (index + 1, v_address, v))
                    continue
                else:
                    v = v_address
            print("   dw[%d] = %s;" % (index, v))
            print("   dw[%d] = %s >> 32;" % (index + 1, v))

class Value(object):
    def __init__(self, attrs):
        self.name = safe_name(attrs["name"])
        self.value = ast.literal_eval(attrs["value"])
        self.dont_use = int(attrs["dont_use"]) != 0 if "dont_use" in attrs else False

class Parser(object):
    def __init__(self, opencl, repack):
        self.instruction = None
        self.structs = {}
        # Set of enum names we've seen.
        self.enums = set()
        self.registers = {}
        self.opencl = opencl
        self.repack = repack

    def gen_prefix(self, name):
        if name[0] == "_":
            return 'GFX%s%s' % (self.gen, name)
        return 'GFX%s_%s' % (self.gen, name)

    def gen_guard(self):
        if self.opencl:
            return self.gen_prefix("{0}_CL_PACK_H".format(self.platform))
        return self.gen_prefix("{0}_PACK_H".format(self.platform))

    def process_item(self, item):
        name = item.tag
        assert name != "genxml"
        attrs = item.attrib

        if name in ("instruction", "struct", "register"):
            if name == "instruction":
                self.instruction = safe_name(attrs["name"])
                self.length_bias = int(attrs["bias"])
            elif name == "struct":
                self.struct = safe_name(attrs["name"])
                self.structs[attrs["name"]] = 1
            elif name == "register":
                self.register = safe_name(attrs["name"])
                self.reg_num = num_from_str(attrs["num"])
                self.registers[attrs["name"]] = 1
            if "length" in attrs:
                self.length = int(attrs["length"])
                size = self.length * 32
            else:
                self.length = None
                size = 0
            self.group = Group(self, None, 0, 1, size)

        elif name == "group":
            dword = int(attrs["dword"])
            offset_bits = int(attrs.get("offset_bits", 0))
            start = dword * 32 + offset_bits


            group = Group(self, self.group,
                          start, int(attrs["count"]), int(attrs["size"]))
            self.group.fields.append(group)
            self.group = group
        elif name == "field":
            self.group.fields.append(Field(self, attrs))
            self.values = []
        elif name == "enum":
            self.values = []
            self.enum = safe_name(attrs["name"])
            self.enums.add(attrs["name"])
            if "prefix" in attrs:
                self.prefix = safe_name(attrs["prefix"])
            else:
                self.prefix = None
        elif name == "value":
            self.values.append(Value(attrs))
        elif name in ("import", "exclude"):
            pass
        else:
            assert False

        for child_item in item:
            self.process_item(child_item)

        if name  == "instruction":
            self.emit_instruction()
            self.instruction = None
            self.group = None
        elif name == "struct":
            self.emit_struct()
            self.struct = None
            self.group = None
        elif name == "register":
            self.emit_register()
            self.register = None
            self.reg_num = None
            self.group = None
        elif name == "group":
            self.group = self.group.parent
        elif name  == "field":
            self.group.fields[-1].values = self.values
        elif name  == "enum":
            self.emit_enum()
            self.enum = None
        elif name in ("import", "exclude", "value"):
            pass
        else:
            assert False

    def emit_template_struct(self, name, group):
        print("struct %s {" % self.gen_prefix(name))
        group.emit_template_struct("")
        print("};\n")

    def emit_pack_function(self, name, group, repack=False):
        name = self.gen_prefix(name)
        if repack:
            if self.opencl:
                print(textwrap.dedent("""\
                static inline __attribute__((always_inline)) void
                %s_repack(__attribute__((unused)) global void * restrict dst,
                        %s__attribute__((unused)) global const uint32_t * origin,
                        %s__attribute__((unused)) private const struct %s * restrict values)
                {""") % (name, ' ' * len(name), ' ' * len(name), name))
            else:
                print(textwrap.dedent("""\
                static inline __attribute__((always_inline)) void
                %s_repack(__attribute__((unused)) __gen_user_data *data,
                        %s__attribute__((unused)) void * restrict dst,
                        %s__attribute__((unused)) global const uint32_t * origin,
                        %s__attribute__((unused)) const struct %s * restrict values)
                {""") % (name, ' ' * len(name), ' ' * len(name), ' ' * len(name), name))
        else:
            if self.opencl:
                print(textwrap.dedent("""\
                static inline __attribute__((always_inline)) void
                %s_pack(__attribute__((unused)) global void * restrict dst,
                      %s__attribute__((unused)) private const struct %s * restrict values)
                {""") % (name, ' ' * len(name), name))
            else:
                print(textwrap.dedent("""\
                static inline __attribute__((always_inline)) void
                %s_pack(__attribute__((unused)) __gen_user_data *data,
                      %s__attribute__((unused)) void * restrict dst,
                      %s__attribute__((unused)) const struct %s * restrict values)
                {""") % (name, ' ' * len(name), ' ' * len(name), name))

        (dwords, length) = group.collect_dwords_and_length(repack)
        if length:
            # Cast dst to make header C++ friendly
            type_name = "global uint32_t *" if self.opencl else "uint32_t * restrict"
            print("   %s dw = (%s) dst;" % (type_name, type_name))

            group.emit_pack_function(dwords, length, repack)

        print("}\n")

    def emit_instruction(self):
        name = self.instruction

        if not self.length is None:
            print('#define %-33s %6d' %
                  (self.gen_prefix(name + "_length"), self.length))
        print('#define %-33s %6d' %
              (self.gen_prefix(name + "_length_bias"), self.length_bias))

        default_fields = []
        for field in self.group.fields:
            if not isinstance(field, Field):
                continue
            if field.default is None:
                continue

            if field.is_builtin_type():
                default_fields.append("   .%-35s = %6d" % (field.name, field.default))
            else:
                # Default values should not apply to structures
                assert field.is_enum_type()
                default_fields.append("   .%-35s = (enum %s) %6d" % (field.name, self.gen_prefix(safe_name(field.type)), field.default))

        if default_fields:
            print('#define %-40s\\' % (self.gen_prefix(name + '_header')))
            print(",  \\\n".join(default_fields))
            print('')

        self.emit_template_struct(self.instruction, self.group)
        self.emit_pack_function(self.instruction, self.group)
        if self.repack:
            self.emit_pack_function(self.instruction, self.group, repack=True)

    def emit_register(self):
        name = self.register
        if not self.reg_num is None:
            print('#define %-33s 0x%04x' %
                  (self.gen_prefix(name + "_num"), self.reg_num))

        if not self.length is None:
            print('#define %-33s %6d' %
                  (self.gen_prefix(name + "_length"), self.length))

        self.emit_template_struct(self.register, self.group)
        self.emit_pack_function(self.register, self.group)

    def emit_struct(self):
        name = self.struct
        if not self.length is None:
            print('#define %-33s %6d' %
                  (self.gen_prefix(name + "_length"), self.length))

        self.emit_template_struct(self.struct, self.group)
        self.emit_pack_function(self.struct, self.group)
        if self.repack:
            self.emit_pack_function(self.struct, self.group, repack=True)

    def emit_enum(self):
        print('enum %s {' % self.gen_prefix(self.enum))
        for value in self.values:
            if self.prefix:
                name = self.prefix + "_" + value.name
            else:
                name = value.name
            print('   %-36s = %6d,' % (name.upper(), value.value))
        print('};\n')

    def emit_genxml(self, genxml):
        root = genxml.et.getroot()
        self.platform = root.attrib["name"]
        self.gen = root.attrib["gen"].replace('.', '')
        print(pack_header % {'license': license, 'platform': self.platform, 'guard': self.gen_guard()})
        for item in root:
            self.process_item(item)
        print('#endif /* %s */' % self.gen_guard())

def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument('xml_source', metavar='XML_SOURCE',
                   help="Input xml file")
    p.add_argument('--engines', nargs='?', type=str, default='render',
                   help="Comma-separated list of engines whose instructions should be parsed (default: %(default)s)")
    p.add_argument('--include-symbols', nargs='?', type=str, action='store',
                   help="List of instruction/structures to generate")
    p.add_argument('--opencl', action='store_true', help="Generate OpenCL code")
    p.add_argument('--repack', action='store_true', help="Emit repacking code")

    pargs = p.parse_args()

    if pargs.engines is None:
        print("No engines specified")
        sys.exit(1)

    return pargs

def main():
    pargs = parse_args()

    engines = set(pargs.engines.split(','))
    valid_engines = [ 'render', 'blitter', 'video', 'compute' ]
    if engines - set(valid_engines):
        print("Invalid engine specified, valid engines are:\n")
        for e in valid_engines:
            print("\t%s" % e)
        sys.exit(1)

    genxml = intel_genxml.GenXml(pargs.xml_source)

    genxml.merge_imported()
    genxml.filter_engines(engines)
    if pargs.include_symbols:
        genxml.filter_symbols(pargs.include_symbols.split(','))
    p = Parser(pargs.opencl, pargs.repack)
    p.emit_genxml(genxml)

if __name__ == '__main__':
    main()



COPYRIGHT=u"""
/* Copyright © 2022 Collabora, Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */
"""

import argparse
import os
import re
from collections import namedtuple
import xml.etree.ElementTree as et

from mako.template import Template

TEMPLATE_H = Template(COPYRIGHT + """\
/* This file generated from ${filename}, don't edit directly. */

#ifndef VK_FORMAT_INFO_H
#define VK_FORMAT_INFO_H

#include <vulkan/vulkan_core.h>

#ifdef __cplusplus
extern "C" {
#endif

enum vk_format_class {
   MESA_VK_FORMAT_CLASS_UNKNOWN,
% for name in format_classes:
   ${to_enum_name('MESA_VK_FORMAT_CLASS_', name)},
% endfor
};

struct vk_format_class_info {
   const VkFormat *formats;
   uint32_t format_count;
};

const struct vk_format_class_info *
vk_format_class_get_info(enum vk_format_class class);

const struct vk_format_class_info *
vk_format_get_class_info(VkFormat format);

#ifdef __cplusplus
}
#endif

#endif
""")

TEMPLATE_C = Template(COPYRIGHT + """
/* This file generated from ${filename}, don't edit directly. */

#include "${header}"

#include "util/macros.h"

#include "vk_format.h"

struct vk_format_info {
   enum vk_format_class class;
};

% for id, ext in extensions.items():
static const struct vk_format_info ext${id}_format_infos[] = {
%   for name, format in ext.formats.items():
   [${format.offset}] = {
      .class = ${to_enum_name('MESA_VK_FORMAT_CLASS_', format.cls)},
   },
%   endfor
};

% endfor
static const struct vk_format_info *
vk_format_get_info(VkFormat format)
{
   uint32_t extnumber =
      format < 1000000000 ? 0 : (((format % 1000000000) / 1000) + 1);
   uint32_t offset = format % 1000;

   switch (extnumber) {
% for id, ext in extensions.items():
   case ${id}:
      assert(offset < ARRAY_SIZE(ext${id}_format_infos));
      return &ext${id}_format_infos[offset];
% endfor
   default:
      UNREACHABLE("Invalid extension");
   }
}

% for clsname, cls in format_classes.items():
%   if len(cls.formats) > 0:
static const VkFormat ${to_enum_name('MESA_VK_FORMAT_CLASS_', clsname).lower() + '_formats'}[] = {
%     for fname in cls.formats:
   ${fname},
%     endfor
%   endif
};

% endfor
static const struct vk_format_class_info class_infos[] = {
% for clsname, cls in format_classes.items():
   [${to_enum_name('MESA_VK_FORMAT_CLASS_', clsname)}] = {
%   if len(cls.formats) > 0:
      .formats = ${to_enum_name('MESA_VK_FORMAT_CLASS_', clsname).lower() + '_formats'},
      .format_count = ARRAY_SIZE(${to_enum_name('MESA_VK_FORMAT_CLASS_', clsname).lower() + '_formats'}),
%   else:
      0
%   endif
   },
% endfor
};

const struct vk_format_class_info *
vk_format_class_get_info(enum vk_format_class class)
{
   assert(class < ARRAY_SIZE(class_infos));
   return &class_infos[class];
}

const struct vk_format_class_info *
vk_format_get_class_info(VkFormat format)
{
    const struct vk_format_info *format_info = vk_format_get_info(format);
    return &class_infos[format_info->class];
}
""")

def to_enum_name(prefix, name):
    return "%s" % prefix + re.sub('([^A-Za-z0-9_])', '_', name).upper()

Format = namedtuple('Format', ['name', 'cls', 'ext', 'offset'])
FormatClass = namedtuple('FormatClass', ['name', 'formats'])
Extension = namedtuple('Extension', ['id', 'formats'])

def get_formats(doc):
    """Extract the formats from the registry."""
    formats = {}

    for fmt in doc.findall('./formats/format'):
        xpath = './/enum[@name="{}"]'.format(fmt.attrib['name'])
        enum = doc.find(xpath)
        ext = None
        if 'extends' in enum.attrib:
            assert(enum.attrib['extends'] == 'VkFormat')
            if 'extnumber' in enum.attrib:
                ext = int(enum.attrib['extnumber'])
            else:
                xpath = xpath + '/..'
                parent = doc.find(xpath)
                while parent != None and ext == None:
                    if parent.tag == 'extension':
                        assert('number' in parent.attrib)
                        ext = parent.attrib['number']
                    xpath = xpath + '/..'
                    parent = doc.find(xpath)
            offset = int(enum.attrib['offset'])
        else:
            ext = 0
            offset = int(enum.attrib['value'])

        assert(ext != None)
        format = Format(fmt.attrib['name'], fmt.attrib['class'], ext, offset)
        formats[format.name] = format

    return formats

def get_formats_from_xml(xml_files):
    formats = {}

    for filename in xml_files:
        doc = et.parse(filename)
        formats.update(get_formats(doc))

    return formats

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--out-c', required=True, help='Output C file.')
    parser.add_argument('--out-h', required=True, help='Output H file.')
    parser.add_argument('--xml',
                        help='Vulkan API XML file.',
                        required=True, action='append', dest='xml_files')
    args = parser.parse_args()

    formats = get_formats_from_xml(args.xml_files)
    classes = {}
    extensions = {}
    for n, f in formats.items():
        if f.cls not in classes:
            classes[f.cls] = FormatClass(f.cls, {})
        classes[f.cls].formats[f.name] = f
        if f.ext not in extensions:
            extensions[f.ext] = Extension(f.cls, {})
        extensions[f.ext].formats[f.name] = f

    assert os.path.dirname(args.out_c) == os.path.dirname(args.out_h)

    environment = {
        'header': os.path.basename(args.out_h),
        'formats': formats,
        'format_classes': classes,
        'extensions': extensions,
        'filename': os.path.basename(__file__),
        'to_enum_name': to_enum_name,
    }

    try:
        with open(args.out_h, 'w', encoding='utf-8') as f:
            guard = os.path.basename(args.out_h).replace('.', '_').upper()
            f.write(TEMPLATE_H.render(guard=guard, **environment))
        with open(args.out_c, 'w', encoding='utf-8') as f:
            f.write(TEMPLATE_C.render(**environment))
    except Exception:
        # In the event there's an error, this imports some helpers from mako
        # to print a useful stack trace and prints it, then exits with
        # status 1, if python is run with debug; otherwise it just raises
        # the exception
        import sys
        from mako import exceptions
        print(exceptions.text_error_template().render(), file=sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    main()



# Copyright © 2019,2021 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import textwrap
import typing

import pytest

if typing.TYPE_CHECKING:
    import aiohttp

# AsyncMock is new in 3.8, so if we're using an older version we need the
# backported version of mock
if sys.version_info >= (3, 8):
    from unittest import mock
else:
    import mock

from .gen_release_notes import *


@pytest.mark.parametrize(
    'current, is_point, expected',
    [
        ('19.2.0', True, '19.2.1'),
        ('19.3.6', True, '19.3.7'),
        ('20.0.0-rc4', False, '20.0.0'),
    ])
def test_next_version(current: str, is_point: bool, expected: str) -> None:
    assert calculate_next_version(current, is_point) == expected


@pytest.mark.parametrize(
    'current, is_point, expected',
    [
        ('19.3.6', True, '19.3.6'),
        ('20.0.0-rc4', False, '19.3.0'),
    ])
def test_previous_version(current: str, is_point: bool, expected: str) -> None:
    assert calculate_previous_version(current, is_point) == expected


@pytest.mark.asyncio
async def test_get_shortlog():
    # Certainly not perfect, but it's something
    version = '19.2.0'
    out = await get_shortlog(version)
    assert out


@pytest.mark.asyncio
async def test_gather_commits():
    # Certainly not perfect, but it's something
    version = '19.2.0'
    out = await gather_commits(version)
    assert out


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'content, bugs',
    [
        # It is important to have the title on a new line, as
        # textwrap.dedent won't work otherwise.

        # Test the `Closes: #N` syntax
        (
            '''\
            A commit

            It has a message in it

            Closes: #1
            ''',
            ['1'],
        ),

        # Test the Full url
        (
            '''\
            A commit with no body

            Closes: https://gitlab.freedesktop.org/mesa/mesa/-/issues/3456
            ''',
            ['3456'],
        ),

        # Test projects that are not mesa
        (
            '''\
            A commit for libdrm

            Closes: https://gitlab.freedesktop.org/mesa/drm/-/3456
            ''',
            [],
        ),
        (
            '''\
            A commit for for something else completely

            Closes: https://github.com/Organization/project/1234
            ''',
            [],
        ),

        # Test  multiple issues on one line
        (
            '''\
            Fix many bugs

            Closes: #1, #2
            ''',
            ['1', '2'],
        ),

        # Test multiple closes
        (
            '''\
            Fix many bugs

            Closes: #1
            Closes: #2
            ''',
            ['1', '2'],
        ),
        (
            '''\
            With long form

            Closes: https://gitlab.freedesktop.org/mesa/mesa/-/issues/3456
            Closes: https://gitlab.freedesktop.org/mesa/mesa/-/issues/3457
            Closes: https://gitlab.freedesktop.org/mesa/mesa/-/issues/3458
            ''',
            ['3456', '3457', '3458'],
        ),
        (
            '''\
            Without /-/

            Closes: https://gitlab.freedesktop.org/mesa/mesa/issues/36
            ''',
            ['36'],
        ),
        (
            '''\
            Ignore merge_requests

            Closes: https://gitlab.freedesktop.org/mesa/mesa/-/merge_requests/20241
            ''',
            [],
        ),
        (
            '''\
            Parse "Fixes:" tag too

            Fixes: https://gitlab.freedesktop.org/mesa/mesa/issues/36
            Fixes: 142565a3bc2
            Fixes: 142565a3bc2 ("docs: do something very useful")
            Fixes: 142565a3bc2 ("docs: fix #1234, have a comma")
            Fixes: https://gitlab.freedesktop.org/mesa/mesa/-/issues/37
            ''',
            ['36', '37'],
        ),
        (
            '''\
            Parse Fixes/Closes in weird cases

            fixes: https://gitlab.freedesktop.org/mesa/mesa/issues/36
            fiXES: https://gitlab.freedesktop.org/mesa/mesa/issues/37
            closes: https://gitlab.freedesktop.org/mesa/mesa/issues/38
            cloSES: https://gitlab.freedesktop.org/mesa/mesa/issues/39
            ''',
            ['36', '37', '38', '39'],
        ),
    ])
async def test_parse_issues(content: str, bugs: typing.List[str]) -> None:
    mock_com = mock.AsyncMock(return_value=(textwrap.dedent(content).encode(), ''))
    mock_p = mock.Mock()
    mock_p.communicate = mock_com
    mock_exec = mock.AsyncMock(return_value=mock_p)

    with mock.patch('bin.gen_release_notes.asyncio.create_subprocess_exec', mock_exec), \
            mock.patch('bin.gen_release_notes.gather_commits', mock.AsyncMock(return_value='sha\n')):
        ids = await parse_issues('1234 not used')
        assert set(ids) == set(bugs)


@pytest.mark.asyncio
async def test_rst_escape():
    out = inliner.quoteInline('foo@bar')
    assert out == 'foo\@bar'


@pytest.mark.asyncio
async def test_gather_bugs_duplicates():
    mock_gc = mock.AsyncMock(return_value='something')
    mock_pi = mock.AsyncMock(return_value=['a', 'b', 'a', 'a', 'c', 'b'])

    async def get_bug(session: 'aiohttp.ClientSession', bug_id: str) -> str:
        return bug_id

    with mock.patch('bin.gen_release_notes.gather_commits', mock_gc), \
            mock.patch('bin.gen_release_notes.parse_issues', mock_pi), \
            mock.patch('bin.gen_release_notes.get_bug', get_bug):
        bugs = await gather_bugs('')

    assert bugs == ['a', 'b', 'c']




copyright = '''
/*
 * Copyright 2009 VMware, Inc.
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * VMWARE AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
'''

import argparse
import itertools
import typing as T

GENERATE, UINT8, UINT16, UINT32 = 'generate', 'uint8', 'uint16', 'uint32'
FIRST, LAST = 'first', 'last'

INTYPES = (GENERATE, UINT8, UINT16, UINT32)
OUTTYPES = (UINT16, UINT32)
PRIMS=('tris',
       'trifan',
       'tristrip',
       'quads',
       'quadstrip',
       'polygon',
       'trisadj',
       'tristripadj')

LONGPRIMS=('MESA_PRIM_TRIANGLES',
           'MESA_PRIM_TRIANGLE_FAN',
           'MESA_PRIM_TRIANGLE_STRIP',
           'MESA_PRIM_QUADS',
           'MESA_PRIM_QUAD_STRIP',
           'MESA_PRIM_POLYGON',
           'MESA_PRIM_TRIANGLES_ADJACENCY',
           'MESA_PRIM_TRIANGLE_STRIP_ADJACENCY')

longprim = dict(zip(PRIMS, LONGPRIMS))
intype_idx = dict(uint8='IN_UINT8', uint16='IN_UINT16', uint32='IN_UINT32')
outtype_idx = dict(uint16='OUT_UINT16', uint32='OUT_UINT32')

def prolog(f: 'T.TextIO'):
    f.write('/* File automatically generated by u_unfilled_gen.py */\n')
    f.write(copyright)
    f.write(r'''

/**
 * @file
 * Functions to translate and generate index lists
 */

#include "indices/u_indices.h"
#include "indices/u_indices_priv.h"
#include "util/compiler.h"
#include "util/u_debug.h"
#include "pipe/p_defines.h"
#include "util/u_memory.h"

static u_generate_func generate_line[OUT_COUNT][PRIM_COUNT];
static u_translate_func translate_line[IN_COUNT][OUT_COUNT][PRIM_COUNT];

''')

def vert( intype, outtype, v0 ):
    if intype == GENERATE:
        return '(' + outtype + '_t)(' + v0 + ')'
    else:
        return '(' + outtype + '_t)in[' + v0 + ']'

def line(f: 'T.TextIO', intype, outtype, ptr, v0, v1 ):
    f.write('      (' + ptr + ')[0] = ' + vert( intype, outtype, v0 ) + ';\n')
    f.write('      (' + ptr + ')[1] = ' + vert( intype, outtype, v1 ) + ';\n')

# XXX: have the opportunity here to avoid over-drawing shared lines in
# tristrips, fans, etc, by integrating this into the calling functions
# and only emitting each line at most once.
#
def do_tri(f: 'T.TextIO', intype, outtype, ptr, v0, v1, v2 ):
    line(f, intype, outtype, ptr, v0, v1 )
    line(f, intype, outtype, ptr + '+2', v1, v2 )
    line(f, intype, outtype, ptr + '+4', v2, v0 )

def do_quad(f: 'T.TextIO', intype, outtype, ptr, v0, v1, v2, v3 ):
    line(f, intype, outtype, ptr, v0, v1 )
    line(f, intype, outtype, ptr + '+2', v1, v2 )
    line(f, intype, outtype, ptr + '+4', v2, v3 )
    line(f, intype, outtype, ptr + '+6', v3, v0 )

def name(intype, outtype, prim):
    if intype == GENERATE:
        return 'generate_' + prim + '_' + outtype
    else:
        return 'translate_' + prim + '_' + intype + '2' + outtype

def preamble(f: 'T.TextIO', intype, outtype, prim):
    f.write('static void ' + name( intype, outtype, prim ) + '(\n')
    if intype != GENERATE:
        f.write('    const void * _in,\n')
    f.write('    unsigned start,\n')
    if intype != GENERATE:
        f.write('    unsigned in_nr,\n')
    f.write('    unsigned out_nr,\n')
    if intype != GENERATE:
        f.write('    unsigned restart_index,\n')
    f.write('    void *_out )\n')
    f.write('{\n')
    if intype != GENERATE:
        f.write('  const ' + intype + '_t *in = (const ' + intype + '_t*)_in;\n')
    f.write('  ' + outtype + '_t *out = (' + outtype + '_t*)_out;\n')
    f.write('  unsigned i, j;\n')

def postamble(f: 'T.TextIO'):
    f.write('}\n')


def tris(f: 'T.TextIO', intype, outtype):
    preamble(f, intype, outtype, prim='tris')
    f.write('  for (i = start, j = 0; j < out_nr; j+=6, i+=3) {\n')
    do_tri(f, intype, outtype, 'out+j',  'i', 'i+1', 'i+2' );
    f.write('   }\n')
    postamble(f)


def tristrip(f: 'T.TextIO', intype, outtype):
    preamble(f, intype, outtype, prim='tristrip')
    f.write('  for (i = start, j = 0; j < out_nr; j+=6, i++) {\n')
    do_tri(f, intype, outtype, 'out+j',  'i', 'i+1/*+(i&1)*/', 'i+2/*-(i&1)*/' );
    f.write('   }\n')
    postamble(f)


def trifan(f: 'T.TextIO', intype, outtype):
    preamble(f, intype, outtype, prim='trifan')
    f.write('  for (i = start, j = 0; j < out_nr; j+=6, i++) {\n')
    do_tri(f, intype, outtype, 'out+j',  '0', 'i+1', 'i+2' );
    f.write('   }\n')
    postamble(f)



def polygon(f: 'T.TextIO', intype, outtype):
    preamble(f, intype, outtype, prim='polygon')
    f.write('  for (i = start, j = 0; j < out_nr; j+=2, i++) {\n')
    line(f, intype, outtype, 'out+j', 'i', '(i+1)%(out_nr/2)\n' )
    f.write('   }\n')
    postamble(f)


def quads(f: 'T.TextIO', intype, outtype):
    preamble(f, intype, outtype, prim='quads')
    f.write('  for (i = start, j = 0; j < out_nr; j+=8, i+=4) {\n')
    do_quad(f, intype, outtype, 'out+j', 'i+0', 'i+1', 'i+2', 'i+3' );
    f.write('   }\n')
    postamble(f)


def quadstrip(f: 'T.TextIO', intype, outtype):
    preamble(f, intype, outtype, prim='quadstrip')
    f.write('  for (i = start, j = 0; j < out_nr; j+=8, i+=2) {\n')
    do_quad(f, intype, outtype, 'out+j', 'i+2', 'i+0', 'i+1', 'i+3' );
    f.write('   }\n')
    postamble(f)


def trisadj(f: 'T.TextIO', intype, outtype):
    preamble(f, intype, outtype, prim='trisadj')
    f.write('  for (i = start, j = 0; j < out_nr; j+=6, i+=6) {\n')
    do_tri(f, intype, outtype, 'out+j',  'i', 'i+2', 'i+4' );
    f.write('   }\n')
    postamble(f)


def tristripadj(f: 'T.TextIO', intype, outtype):
    preamble(f, intype, outtype, prim='tristripadj')
    f.write('  for (i = start, j = 0; j < out_nr; j+=6, i+=2) {\n')
    do_tri(f, intype, outtype, 'out+j',  'i', 'i+2', 'i+4' );
    f.write('   }\n')
    postamble(f)


def emit_funcs(f: 'T.TextIO'):
    for intype, outtype in itertools.product(INTYPES, OUTTYPES):
        tris(f, intype, outtype)
        tristrip(f, intype, outtype)
        trifan(f, intype, outtype)
        quads(f, intype, outtype)
        quadstrip(f, intype, outtype)
        polygon(f, intype, outtype)
        trisadj(f, intype, outtype)
        tristripadj(f, intype, outtype)

def init(f: 'T.TextIO', intype, outtype, prim):
    if intype == GENERATE:
        f.write('generate_line[' +
                outtype_idx[outtype] +
                '][' + longprim[prim] +
                '] = ' + name( intype, outtype, prim ) + ';\n')
    else:
        f.write('translate_line[' +
                intype_idx[intype] +
                '][' + outtype_idx[outtype] +
                '][' + longprim[prim] +
                '] = ' + name( intype, outtype, prim ) + ';\n')


def emit_all_inits(f: 'T.TextIO'):
    for intype, outtype, prim in itertools.product(INTYPES, OUTTYPES, PRIMS):
        init(f, intype, outtype, prim)

def emit_init(f: 'T.TextIO'):
    f.write('void u_unfilled_init( void )\n')
    f.write('{\n')
    f.write('  static int firsttime = 1;\n')
    f.write('  if (!firsttime) return;\n')
    f.write('  firsttime = 0;\n')
    emit_all_inits(f)
    f.write('}\n')




def epilog(f: 'T.TextIO'):
    f.write('#include "indices/u_unfilled_indices.c"\n')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('output')
    args = parser.parse_args()

    with open(args.output, 'w') as f:
        prolog(f)
        emit_funcs(f)
        emit_init(f)
        epilog(f)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright © 2023 Collabora Ltd.
# Authors:
#   Helen Koike <helen.koike@collabora.com>
#
# For the dependencies, see the requirements.txt
# SPDX-License-Identifier: MIT


import argparse
import logging as log
import os
import re
import traceback
from datetime import datetime, timedelta
from typing import Any, Dict

import gitlab
import pytz
from ci_gantt_chart import generate_gantt_chart
from gitlab import Gitlab
from gitlab.base import RESTObject
from gitlab.v4.objects import Project, ProjectPipeline
from gitlab_common import (GITLAB_URL, get_gitlab_pipeline_from_url,
                           get_token_from_default_dir, read_token)


class MockGanttExit(Exception):
    pass


LAST_MARGE_EVENT_FILE = os.path.expanduser("~/.config/last_marge_event")


def read_last_event_date_from_file() -> str:
    try:
        with open(LAST_MARGE_EVENT_FILE, "r") as f:
            last_event_date = f.read().strip()
    except FileNotFoundError:
        # 3 days ago
        last_event_date = (datetime.now() - timedelta(days=3)).isoformat()
    return last_event_date


def pretty_time(time_str: str) -> str:
    """Pretty print time"""
    local_timezone = datetime.now().astimezone().tzinfo

    time_d = datetime.fromisoformat(time_str.replace("Z", "+00:00")).astimezone(
        local_timezone
    )
    return f'{time_str} ({time_d.strftime("%d %b %Y %Hh%Mm%Ss")} {local_timezone})'


def compose_message(file_name: str, attachment_url: str) -> str:
    return f"""
[{file_name}]({attachment_url})

<details>
<summary>more info</summary>

This message was generated by the ci_post_gantt.py script, which is running on a server at Collabora.
</details>
"""


def gitlab_upload_file_get_url(gl: Gitlab, project_id: str, filepath: str) -> str:
    project: Project = gl.projects.get(project_id)
    uploaded_file: Dict[str, Any] = project.upload(filepath, filepath=filepath)
    return uploaded_file["url"]


def gitlab_post_reply_to_note(gl: Gitlab, event: RESTObject, reply_message: str):
    """
    Post a reply to a note in thread based on a GitLab event.

    :param gl: The GitLab connection instance.
    :param event: The event object containing the note details.
    :param reply_message: The reply message.
    """
    try:
        note_id = event.target_id
        merge_request_iid = event.note["noteable_iid"]

        project = gl.projects.get(event.project_id)
        merge_request = project.mergerequests.get(merge_request_iid)

        # Find the discussion to which the note belongs
        discussions = merge_request.discussions.list(iterator=True)
        target_discussion = next(
            (
                d
                for d in discussions
                if any(n["id"] == note_id for n in d.attributes["notes"])
            ),
            None,
        )

        if target_discussion is None:
            raise ValueError("Discussion for the note not found.")

        # Add a reply to the discussion
        reply = target_discussion.notes.create({"body": reply_message})
        return reply

    except gitlab.exceptions.GitlabError as e:
        log.error(f"Failed to post a reply to '{event.note['body']}': {e}")
        return None


def main(
    token: str | None,
    since: str | None,
    marge_user_id: int = 9716,
    project_ids: list[int] = [176],
    ci_timeout: float = 60,
):
    log.basicConfig(level=log.INFO)
    if token is None:
        token = get_token_from_default_dir()

    token = read_token(token)
    gl = Gitlab(url=GITLAB_URL, private_token=token, retry_transient_errors=True)

    user = gl.users.get(marge_user_id)
    last_event_at = since if since else read_last_event_date_from_file()

    log.info(f"Retrieving Marge messages since {pretty_time(last_event_at)}\n")

    # the "after" only considers the "2023-10-24" part, it doesn't consider the time
    events = user.events.list(
        all=True,
        target_type="note",
        after=(datetime.now() - timedelta(days=3)).isoformat(),
        sort="asc",
    )

    last_event_at_date = datetime.fromisoformat(
        last_event_at.replace("Z", "+00:00")
    ).replace(tzinfo=pytz.UTC)

    for event in events:
        if event.project_id not in project_ids:
            continue
        created_at_date = datetime.fromisoformat(
            event.created_at.replace("Z", "+00:00")
        ).replace(tzinfo=pytz.UTC)
        if created_at_date <= last_event_at_date:
            continue
        last_event_at = event.created_at

        escaped_gitlab_url = re.escape(GITLAB_URL)
        match = re.search(rf"{escaped_gitlab_url}/[^\s<]+", event.note["body"])

        if match:
            try:
                log.info(f"Found message: {event.note['body']}")
                pipeline_url = match.group(0)[:-1]
                pipeline: ProjectPipeline
                pipeline, _ = get_gitlab_pipeline_from_url(gl, pipeline_url)
                log.info("Generating gantt chart...")
                fig = generate_gantt_chart(pipeline, ci_timeout)
                file_name = f"{str(pipeline.id)}-Gantt.html"
                fig.write_html(file_name)
                log.info("Uploading gantt file...")
                file_url = gitlab_upload_file_get_url(gl, event.project_id, file_name)
                log.info("Posting reply ...")
                message = compose_message(file_name, file_url)
                gitlab_post_reply_to_note(gl, event, message)
            except MockGanttExit:
                pass  # Allow tests to exit early without printing a traceback
            except Exception as e:
                log.info(f"Failed to generate gantt chart, not posting reply.{e}")
                traceback.print_exc()

        if not since:
            log.info(
                f"Updating last event date to {pretty_time(last_event_at)} on {LAST_MARGE_EVENT_FILE}\n"
            )
            with open(LAST_MARGE_EVENT_FILE, "w") as f:
                f.write(last_event_at)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Monitor rejected pipelines by Marge.")
    parser.add_argument(
        "--token",
        metavar="token",
        type=str,
        default=None,
        help="force GitLab token, otherwise it's read from ~/.config/gitlab-token",
    )
    parser.add_argument(
        "--since",
        metavar="since",
        type=str,
        default=None,
        help="consider only events after this date (ISO format), otherwise it's read from ~/.config/last_marge_event",
    )
    parser.add_argument(
        "--marge-user-id",
        metavar="marge_user_id",
        type=int,
        default=9716,  # default https://gitlab.freedesktop.org/users/marge-bot/activity
        help="GitLab user ID for marge-bot, defaults to 9716",
    )
    parser.add_argument(
        "--project-id",
        metavar="project_id",
        type=int,
        nargs="+",
        default=[176],  # default is the mesa/mesa project id
        help="GitLab project id(s) to analyze. Defaults to 176 i.e. mesa/mesa.",
    )
    parser.add_argument(
        "--ci-timeout",
        metavar="ci_timeout",
        type=float,
        default=60,
        help="Time that marge-bot will wait for ci to finish. Defaults to one hour.",
    )
    args = parser.parse_args()
    main(args.token, args.since, args.marge_user_id, args.project_id, args.ci_timeout)



# Copyright 2018 Google LLC
# SPDX-License-Identifier: MIT

from .wrapperdefs import VulkanWrapperGenerator


class VulkanExtensionStructureType(VulkanWrapperGenerator):
    def __init__(self, extensionName: str, module, typeInfo):
        super().__init__(module, typeInfo)
        self._extensionName = extensionName

    def onGenGroup(self, groupinfo, groupName, alias=None):
        super().onGenGroup(groupinfo, groupName, alias)
        elem = groupinfo.elem
        if (not elem.get('type') == 'enum'):
            return
        if (not elem.get('name') == 'VkStructureType'):
            return
        extensionEnumFactoryMacro = f'{self._extensionName.upper()}_ENUM'
        for enum in elem.findall(f"enum[@extname='{self._extensionName}']"):
            name = enum.get('name')
            offset = enum.get('offset')
            self.module.appendHeader(
                f"#define {name} {extensionEnumFactoryMacro}(VkStructureType, {offset})\n")


class VulkanGfxstreamStructureType(VulkanExtensionStructureType):
    def __init__(self, module, typeInfo):
        super().__init__('VK_GOOGLE_gfxstream', module, typeInfo)


class VulkanAndroidNativeBufferStructureType(VulkanExtensionStructureType):
    def __init__(self, module, typeInfo):
        super().__init__('VK_ANDROID_native_buffer', module, typeInfo)



COPYRIGHT=u"""
/* Copyright © 2023 Collabora, Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */
"""

import argparse
import os
import textwrap
import xml.etree.ElementTree as et

from mako.template import Template
from vk_extensions import get_api_list

TEMPLATE_C = Template(COPYRIGHT + """\
#include "vk_synchronization.h"

VkPipelineStageFlags2
vk_expand_pipeline_stage_flags2(VkPipelineStageFlags2 stages)
{
% for (group_stage, stages) in group_stages.items():
    if (stages & ${group_stage})
        stages |= ${' |\\n                  '.join(stages)};

% endfor
    if (stages & VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT) {
% for (guard, stage) in all_commands_stages:
% if guard is not None:
#ifdef ${guard}
% endif
        stages |= ${stage};
% if guard is not None:
#endif
% endif
% endfor
    }

    return stages;
}

VkAccessFlags2
vk_read_access2_for_pipeline_stage_flags2(VkPipelineStageFlags2 stages)
{
    VkAccessFlags2 access = 0;

% for ((guard, stages), access) in stages_read_access.items():
% if guard is not None:
#ifdef ${guard}
% endif
    if (stages & (${' |\\n                  '.join(stages)}))
        access |= ${' |\\n                  '.join(access)};
% if guard is not None:
#endif
% endif

% endfor
    return access;
}

VkAccessFlags2
vk_write_access2_for_pipeline_stage_flags2(VkPipelineStageFlags2 stages)
{
    VkAccessFlags2 access = 0;

% for ((guard, stages), access) in stages_write_access.items():
% if guard is not None:
#ifdef ${guard}
% endif
    if (stages & (${' |\\n                  '.join(stages)}))
        access |= ${' |\\n                  '.join(access)};
% if guard is not None:
#endif
% endif

% endfor
    return access;
}
""")

def get_guards(xml, api):
    guards = {}
    for ext_elem in xml.findall('./extensions/extension'):
        supported = get_api_list(ext_elem.attrib['supported'])
        if api not in supported:
            continue

        for enum in ext_elem.findall('./require/enum[@extends]'):
            if enum.attrib['extends'] not in ('VkPipelineStageFlagBits2',
                                              'VkAccessFlagBits2'):
                continue

            if 'protect' not in enum.attrib:
                continue

            name = enum.attrib['name']
            guard = enum.attrib['protect']
            guards[name] = guard

    return guards

def get_all_commands_stages(xml, guards):
    stages = []
    for stage in xml.findall('./sync/syncstage'):
        stage_name = stage.attrib['name']

        exclude = [
            # This isn't a real stage
            'VK_PIPELINE_STAGE_2_NONE',

            # These are real stages but they're a bit weird to include in
            # ALL_COMMANDS because they're context-dependent, depending on
            # whether they're part of srcStagesMask or dstStagesMask.
            #
            # We could avoid all grouped stages but then if someone adds
            # another group later, the behavior of this function may change in
            # a backwards-compatible way.  Also, the other ones aren't really
            # hurting anything if we add them in.
            'VK_PIPELINE_STAGE_2_TOP_OF_PIPE_BIT',
            'VK_PIPELINE_STAGE_2_BOTTOM_OF_PIPE_BIT',

            # This is all COMMANDS, not host.
            'VK_PIPELINE_STAGE_2_HOST_BIT',
        ]
        if stage_name in exclude:
            continue

        guard = guards.get(stage_name, None)
        stages.append((guard, stage_name))

    return stages

def get_group_stages(xml):
    group_stages = {}
    for stage in xml.findall('./sync/syncstage'):
        name = stage.attrib['name']
        equiv = stage.find('./syncequivalent')
        if equiv is not None:
            stages = equiv.attrib['stage'].split(',')
            group_stages[name] = stages

    return group_stages

def access_is_read(name):
    if 'READ' in name:
        assert 'WRITE' not in name
        return True
    elif 'WRITE' in name:
        return False
    else:
        print(name)
        assert False, "Invalid access bit name"

def get_stages_access(xml, read, guards, all_commands_stages, group_stages):
    stages_access = {}
    for access in xml.findall('./sync/syncaccess'):
        access_name = access.attrib['name']
        if access_name == 'VK_ACCESS_2_NONE':
            continue

        if access_is_read(access_name) != read:
            continue

        guard = guards.get(access_name, None)
        support = access.find('./syncsupport')
        if support is not None:
            stages = support.attrib['stage'].split(',')

            for stage in stages:
                if (guard, stage) in all_commands_stages:
                    stages.append('VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT')
                    stages.append('VK_PIPELINE_STAGE_2_TOP_OF_PIPE_BIT' if read else 'VK_PIPELINE_STAGE_2_BOTTOM_OF_PIPE_BIT')
                    break

            for (group, equiv) in group_stages.items():
                for stage in stages:
                    if stage in equiv:
                        stages.append(group)
                        break


            stages.sort()
            key = (guard, tuple(stages))
            if key in stages_access:
                stages_access[key].append(access_name)
            else:
                stages_access[key] = [access_name]

    return stages_access

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--beta', required=True, help='Enable beta extensions.')
    parser.add_argument('--xml', required=True, help='Vulkan API XML file')
    parser.add_argument('--out-c', required=True, help='Output C file.')
    args = parser.parse_args()

    xml = et.parse(args.xml);

    guards = get_guards(xml, 'vulkan')
    all_commands_stages = get_all_commands_stages(xml, guards)
    group_stages = get_group_stages(xml)

    environment = {
        'all_commands_stages': all_commands_stages,
        'group_stages': group_stages,
        'stages_read_access': get_stages_access(xml, True, guards, all_commands_stages, group_stages),
        'stages_write_access': get_stages_access(xml, False, guards, all_commands_stages, group_stages),
    }

    try:
        with open(args.out_c, 'w', encoding='utf-8') as f:
            f.write(TEMPLATE_C.render(**environment))
    except Exception:
        # In the event there's an error, this imports some helpers from mako
        # to print a useful stack trace and prints it, then exits with
        # status 1, if python is run with debug; otherwise it just raises
        # the exception
        import sys
        from mako import exceptions
        print(exceptions.text_error_template().render(), file=sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    main()
